#include <linux/kernel.h>
#include <asm/atomic.h>
#include <cust_leds.h>
#include <cust_leds_def.h>
#include <mach/mt_reg_base.h>
#include <mach/mt_clkmgr.h>
#include <mach/mt_gpio.h>
#include <cust_gpio_usage.h>
#include "ddp_reg.h"
#include "ddp_pwm.h"
#include "ddp_path.h"


#define PWM_DEFAULT_DIV_VALUE 0x0

#define PWM_ERR(fmt, arg...) printk(KERN_ERR "[PWM] " fmt "\n", ##arg)
#define PWM_NOTICE(fmt, arg...) printk(KERN_NOTICE "[PWM] " fmt "\n", ##arg)
#define PWM_MSG(fmt, arg...) printk(KERN_DEBUG "[PWM] " fmt "\n", ##arg)

#define pwm_get_reg_base(id) (DISPSYS_PWM0_BASE)

#define index_of_pwm(id) (0)


static disp_pwm_id_t g_pwm_main_id = DISP_PWM0;
static atomic_t g_pwm_backlight[1] = { ATOMIC_INIT(-1) };
static volatile int g_pwm_max_backlight[1] = { 1023 };
static ddp_module_notify g_ddp_notify = NULL;

static int disp_pwm_config_init(DISP_MODULE_ENUM module, disp_ddp_path_config* pConfig, void* cmdq)
{
    struct cust_mt65xx_led *cust_led_list;
    struct cust_mt65xx_led *cust;
    struct PWM_config *config_data;
    unsigned int pwm_div;
    disp_pwm_id_t id = DISP_PWM0;
    unsigned long reg_base = pwm_get_reg_base(id);
    int index = index_of_pwm(id);

    pwm_div = PWM_DEFAULT_DIV_VALUE;
#if 0    
    cust_led_list = get_cust_led_list();
    if (cust_led_list)
    {
        /* WARNING: may overflow if MT65XX_LED_TYPE_LCD not configured properly */
        cust = &cust_led_list[MT65XX_LED_TYPE_LCD];
        if ((strcmp(cust->name,"lcd-backlight") == 0) && (cust->mode == MT65XX_LED_MODE_CUST_BLS_PWM))
        {
            config_data = &cust->config_data;
            if (config_data->clock_source >= 0 && config_data->clock_source <= 3)
            {
                unsigned int regVal = DISP_REG_GET(CLK_CFG_1);
                clkmux_sel(MT_MUX_PWM, config_data->clock_source, "DISP_PWM");
                PWM_MSG("disp_pwm_init : CLK_CFG_1 0x%x => 0x%x", regVal, DISP_REG_GET(CLK_CFG_1));
            }
            /* Some backlight chip/PMIC(e.g. MT6332) only accept slower clock */
            pwm_div = (config_data->div == 0) ? PWM_DEFAULT_DIV_VALUE : config_data->div;
            pwm_div &= 0x3FF;
            PWM_MSG("disp_pwm_init : PWM config data (%d,%d)", config_data->clock_source, config_data->div);
        }
    }
	
#endif

    atomic_set(&g_pwm_backlight[index], -1);

    /* We don't enable PWM until we really need */
    DISP_REG_MASK(cmdq, reg_base + DISP_PWM_CON_0_OFF, pwm_div << 16, (0x3ff << 16));

    DISP_REG_MASK(cmdq, reg_base + DISP_PWM_CON_1_OFF, 1023, 0x3ff); /* 1024 levels */
    /* We don't init the backlight here until AAL/Android give */

    return 0;
}


static int disp_pwm_config(DISP_MODULE_ENUM module, disp_ddp_path_config* pConfig, void* cmdq)
{
    int ret = 0;

    if (pConfig->dst_dirty)
        ret |= disp_pwm_config_init(module, pConfig, cmdq);

    return ret;
}


static void disp_pwm_trigger_refresh(disp_pwm_id_t id)
{
    if (g_ddp_notify != NULL)
        g_ddp_notify(DISP_MODULE_PWM0, DISP_PATH_EVENT_TRIGGER);
}


/* Set the PWM which acts by default (e.g. ddp_bls_set_backlight) */
void disp_pwm_set_main(disp_pwm_id_t main)
{
    g_pwm_main_id = main;
}


disp_pwm_id_t disp_pwm_get_main(void)
{
    return g_pwm_main_id;
}


int disp_pwm_is_enabled(disp_pwm_id_t id)
{
    unsigned long reg_base = pwm_get_reg_base(id);
    return (DISP_REG_GET(reg_base + DISP_PWM_EN_OFF) & 0x1);
}


static void disp_pwm_set_drverIC_en(disp_pwm_id_t id, int enabled)
{
#ifdef GPIO_LCM_LED_EN
#ifndef CONFIG_MTK_FPGA
    if (id == DISP_PWM0) {
        mt_set_gpio_mode(GPIO_LCM_LED_EN, GPIO_MODE_00);
        mt_set_gpio_dir(GPIO_LCM_LED_EN, GPIO_DIR_OUT);

        if (enabled)
            mt_set_gpio_out(GPIO_LCM_LED_EN, GPIO_OUT_ONE);
        else
            mt_set_gpio_out(GPIO_LCM_LED_EN, GPIO_OUT_ZERO);
    }
#endif
#endif
}


static void disp_pwm_set_enabled(cmdqRecHandle cmdq, disp_pwm_id_t id, int enabled)
{
    unsigned long reg_base = pwm_get_reg_base(id);

    if (enabled) {
        if (!disp_pwm_is_enabled(id)) {
            DISP_REG_MASK(cmdq, reg_base + DISP_PWM_EN_OFF, 0x1, 0x1);

            disp_pwm_set_drverIC_en(id, enabled);
        }
    } else {
        DISP_REG_MASK(cmdq, reg_base + DISP_PWM_EN_OFF, 0x0, 0x1);
        disp_pwm_set_drverIC_en(id, enabled);
    }
}


int disp_bls_set_max_backlight(unsigned int level_1024)
{
    return disp_pwm_set_max_backlight(disp_pwm_get_main(), level_1024);
}


int disp_pwm_set_max_backlight(disp_pwm_id_t id, unsigned int level_1024)
{
    int index;

    if ((DISP_PWM_ALL & id) == 0) {
        PWM_ERR("[ERROR] disp_pwm_set_backlight: invalid PWM ID = 0x%x", id);
        return -EFAULT;
    }
    
    index = index_of_pwm(id);
    g_pwm_max_backlight[index] = level_1024;

    if (level_1024 < atomic_read(&g_pwm_backlight[index]))
        disp_pwm_set_backlight(id, level_1024);

    return 0;
}


int disp_pwm_get_max_backlight(disp_pwm_id_t id)
{
	int index = index_of_pwm(id);
	return g_pwm_max_backlight[index];
}


/* For backward compatible */
int disp_bls_set_backlight(int level_1024)
{
    return disp_pwm_set_backlight(disp_pwm_get_main(), level_1024);
}


/*
 * If you want to re-map the backlight level from user space to
 * the real level of hardware output, please modify here.
 *
 * Inputs:
 *  id          - DISP_PWM0 / DISP_PWM1
 *  level_1024  - Backlight value in [0, 1023]
 * Returns:
 *  PWM duty in [0, 1023]
 */
static int disp_pwm_level_remap(disp_pwm_id_t id, int level_1024)
{
    return level_1024;
}

int disp_pwm_set_backlight(disp_pwm_id_t id, int level_1024)
{
    int ret;

#ifdef MTK_DISP_IDLE_LP
    disp_exit_idle_ex("disp_pwm_set_backlight");
#endif

    /* Always write registers by CPU */
    ret = disp_pwm_set_backlight_cmdq(id, level_1024, NULL);

    if (ret >= 0)
        disp_pwm_trigger_refresh(id);

    return 0;
}


static volatile int g_pwm_duplicate_count = 0;

int disp_pwm_set_backlight_cmdq(disp_pwm_id_t id, int level_1024, void *cmdq)
{
    unsigned long reg_base;
    int old_pwm;
    int index;
    int abs_diff;

    if ((DISP_PWM_ALL & id) == 0) {
        PWM_ERR("[ERROR] disp_pwm_set_backlight_cmdq: invalid PWM ID = 0x%x", id);
        return -EFAULT;
    }

    index = index_of_pwm(id);

    old_pwm = atomic_xchg(&g_pwm_backlight[index], level_1024);
    if (old_pwm != level_1024) {
        abs_diff = level_1024 - old_pwm;
        if (abs_diff < 0) abs_diff = -abs_diff;
        
        if (old_pwm == 0 || level_1024 == 0 || abs_diff > 64) {
            /* To be printed in UART log */
            PWM_NOTICE("disp_pwm_set_backlight_cmdq(id = 0x%x, level_1024 = %d), old = %d", id, level_1024, old_pwm);
        }

        if (level_1024 > g_pwm_max_backlight[index]) {
            level_1024 = g_pwm_max_backlight[index];
        } else if (level_1024 < 0) {
            level_1024 = 0;
        }

        level_1024 = disp_pwm_level_remap(id, level_1024);

        reg_base = pwm_get_reg_base(id);
        DISP_REG_MASK(cmdq, reg_base + DISP_PWM_CON_1_OFF, level_1024 << 16, 0x1fff << 16);

        if (level_1024 > 0) {
            disp_pwm_set_enabled(cmdq, id, 1);
        } else {
            disp_pwm_set_enabled(cmdq, id, 0); /* To save power */
        }

        DISP_REG_MASK(cmdq, reg_base + DISP_PWM_COMMIT_OFF, 1, ~0);
        DISP_REG_MASK(cmdq, reg_base + DISP_PWM_COMMIT_OFF, 0, ~0);

        g_pwm_duplicate_count = 0;
    } else {
        g_pwm_duplicate_count = (g_pwm_duplicate_count + 1) & 63;
    }

    return 0;
}


static int ddp_pwm_power_on(DISP_MODULE_ENUM module, void *handle)
{
    if (module == DISP_MODULE_PWM0) {
       	enable_clock(MT_CG_PWM_MM_SW_CG, "PWM");
    }
    
    return 0;
}

static int ddp_pwm_power_off(DISP_MODULE_ENUM module, void *handle)
{
    if (module == DISP_MODULE_PWM0) {
        atomic_set(&g_pwm_backlight[0], 0);
       	disable_clock(MT_CG_PWM_MM_SW_CG, "PWM");
    }

    return 0;
}


static int ddp_pwm_init(DISP_MODULE_ENUM module, void *cmq_handle)
{
    ddp_pwm_power_on(module, cmq_handle);
    return 0;
}

static int ddp_pwm_set_listener(DISP_MODULE_ENUM module, ddp_module_notify notify)
{
    g_ddp_notify = notify;
    return 0;
}



DDP_MODULE_DRIVER ddp_driver_pwm =
{
    .init           = ddp_pwm_init,
    .config         = disp_pwm_config,
    .power_on       = ddp_pwm_power_on,
    .power_off      = ddp_pwm_power_off,
    .set_listener   = ddp_pwm_set_listener,
};


// ----------------------------------------------------------------------
// Test code
// Following is only for PWM functional test, not normal code
// Will not be linked into user build.
// ----------------------------------------------------------------------

static void disp_pwm_test_source(const char *cmd)
{
	unsigned long reg_base = pwm_get_reg_base(DISP_PWM0);
	int sel = (cmd[0] - '0') & 0x3;
	DISP_REG_MASK(NULL, reg_base + DISP_PWM_CON_0_OFF, (sel << 4), (0x3 << 4));
}


static void disp_pwm_test_grad(const char *cmd)
{
	const unsigned long reg_grad = pwm_get_reg_base(DISP_PWM0) + 0x18;

	switch (cmd[0]) {
	case 'H':
		DISP_REG_SET(NULL, reg_grad, (1 << 16) | (1 << 8) | 1);
		disp_pwm_set_backlight(DISP_PWM0, 1023);
		break;
		
	case 'L':
		DISP_REG_SET(NULL, reg_grad, (1 << 16) | (1 << 8) | 1);
		disp_pwm_set_backlight(DISP_PWM0, 40);
		break;
		
	default:
		DISP_REG_SET(NULL, reg_grad, 0);
		disp_pwm_set_backlight(DISP_PWM0, 512);
		break;
	}
}


static void disp_pwm_test_div(const char *cmd)
{
	const unsigned long reg_base = pwm_get_reg_base(DISP_PWM0);

	int div = cmd[0] - '0';
	if (div > 5)
		div = 5;

	DISP_REG_MASK(NULL, reg_base + DISP_PWM_CON_0_OFF, div << 16, (0x3ff << 16));
	disp_pwm_set_backlight(DISP_PWM0, 256 + div); // to be applied
}


static void disp_pwm_enable_debug(const char *cmd)
{
	const unsigned long reg_base = pwm_get_reg_base(DISP_PWM0);

	if (cmd[0] == '1') {
		DISP_REG_SET(NULL, reg_base + 0x20, 3);
	} else {
		DISP_REG_SET(NULL, reg_base + 0x20, 0);
	}
}


static void disp_pwm_test_pin_mux(void)
{
//#ifndef CONFIG_MTK_FPGA
	const unsigned long reg_base = pwm_get_reg_base(DISP_PWM1);
#if 1
	mt_set_gpio_mode(GPIO66, GPIO_MODE_01);  // For DVT PIN MUX verification only, not normal path
	mt_set_gpio_dir(GPIO66, GPIO_DIR_OUT);	 // For DVT PIN MUX verification only, not normal path

	enable_clock(MT_CG_PWM_MM_SW_CG, "PWM");
#else
	mt_set_gpio_mode(GPIO55, GPIO_MODE_07);  // For DVT PIN MUX verification only, not normal path
	mt_set_gpio_dir(GPIO55, GPIO_DIR_OUT);   // For DVT PIN MUX verification only, not normal path

	mt_set_gpio_mode(GPIO69, GPIO_MODE_01); // For DVT PIN MUX verification only, not normal path
	mt_set_gpio_dir(GPIO69, GPIO_DIR_OUT);  // For DVT PIN MUX verification only, not normal path

	mt_set_gpio_mode(GPIO129, GPIO_MODE_02); // For DVT PIN MUX verification only, not normal path
	mt_set_gpio_dir(GPIO129, GPIO_DIR_OUT);  // For DVT PIN MUX verification only, not normal path

	enable_clock(MT_CG_PERI_DISP_PWM, "PWM");
#endif
	DISP_REG_MASK(NULL, reg_base + DISP_PWM_CON_1_OFF, 512 << 16, 0x1fff << 16);
	DISP_REG_MASK(NULL, reg_base + DISP_PWM_EN_OFF, 0x1, 0x1);
	DISP_REG_SET(NULL, reg_base + 0x20, 3);
	DISP_REG_MASK(NULL, reg_base + DISP_PWM_COMMIT_OFF, 1, ~0);
	DISP_REG_MASK(NULL, reg_base + DISP_PWM_COMMIT_OFF, 0, ~0);
//#endif
}


static void disp_pwm_dump(void)
{
	const unsigned long reg_base = pwm_get_reg_base(DISP_PWM0);
	int offset;

	PWM_NOTICE("[DUMP] Base = 0x%lx", reg_base);
	for (offset = 0; offset <= 0x28; offset += 4) {
		unsigned int val = DISP_REG_GET(reg_base + offset);
		PWM_NOTICE("[DUMP] [+0x%02x] = 0x%08x", offset, val);
	}
}

static int disp_pwm_parse_triple(const char *cmd, unsigned long *offset, unsigned int *value, unsigned int *mask)
{
    int count = 0;
    char *next = (char*)cmd;

    *value = 0;
    *mask = 0;
    *offset = (unsigned long)simple_strtoul(next, &next, 0);

    count++;

    if (*next == ',')
        next++;

    *value = (unsigned int)simple_strtoul(next, &next, 0);
    count++;

    if (*next == ',')
        next++;

    *mask = (unsigned int)simple_strtoul(next, &next, 0);
    count++;

    return count;
}

static void disp_pwm_set_reg(const char *cmd)
{
	const unsigned long reg_base = pwm_get_reg_base(DISP_PWM0);
	unsigned long offset;
    unsigned int value, mask;
	int count = disp_pwm_parse_triple(cmd, &offset, &value, &mask);
        if (count == 3) {
            DISP_REG_MASK(NULL, reg_base + offset, value, mask);
        } else if (count == 2) {
            DISP_REG_SET(NULL, reg_base + offset, value);
            mask = 0xffffffff;
        }
        if (count >= 2) {
        }
}

void disp_pwm_test(const char *cmd, char *debug_output)
{
	debug_output[0] = '\0';

	if (strncmp(cmd, "src:", 4) == 0) {
		disp_pwm_test_source(cmd + 4);
	} else if (strncmp(cmd, "div:", 4) == 0) {
		disp_pwm_test_div(cmd + 4);
	} else if (strncmp(cmd, "set:", 4) == 0) {
		disp_pwm_set_reg(cmd + 4);
	} else if (strncmp(cmd, "grad:", 5) == 0) {
		disp_pwm_test_grad(cmd + 5);
	} else if (strncmp(cmd, "dbg:", 4) == 0) {
		disp_pwm_enable_debug(cmd + 4);
	} else if (strncmp(cmd, "dump", 4) == 0) {
		disp_pwm_dump();
	} else if (strncmp(cmd, "pinmux", 6) == 0) {
		disp_pwm_test_pin_mux();
	}
}
