#include <linux/xlog.h>

#include <mach/upmu_common.h>
#include <mach/upmu_hw.h>
#include <linux/delay.h>

#include <mach/battery_meter_hal.h>
#include <cust_battery_meter.h>
     
//============================================================ //
//define
//============================================================ //
#define STATUS_OK    0
#define STATUS_UNSUPPORTED    -1
#define VOLTAGE_FULL_RANGE     1800
#define ADC_PRECISE           32768  // 15 bits

//============================================================ //
//global variable
//============================================================ //
kal_int32 chip_diff_trim_value_4_0 = 0;
kal_int32 chip_diff_trim_value = 0; // unit = 0.1

kal_int32 g_hw_ocv_tune_value = 8;

kal_bool g_fg_is_charging = 0;

//============================================================ //
//function prototype
//============================================================ //

//============================================================ //
//extern variable
//============================================================ //
 
//============================================================ //
//extern function
//============================================================ //
extern kal_uint32 PMIC_IMM_GetOneChannelValue(mt6350_adc_ch_list_enum dwChannel, int deCount, int trimd);

extern kal_uint32 upmu_get_reg_value(kal_uint32 reg);
extern int IMM_GetOneChannelValue(int dwChannel, int data[4], int* rawdata);
extern int IMM_IsAdcInitReady(void);

//============================================================ // 
void get_hw_chip_diff_trim_value(void)
{
#if defined(CONFIG_POWER_EXT)
#else
    kal_int32 reg_val_1 = 0;
    kal_int32 reg_val_2 = 0;
    
    #if 1
    reg_val_1 = upmu_get_reg_value(0x01C4);
    reg_val_1 = (reg_val_1 & 0xE000) >> 13;
    
    reg_val_2 = upmu_get_reg_value(0x01C6);
    reg_val_2 = (reg_val_2 & 0x0003);

    chip_diff_trim_value_4_0 = reg_val_1 | (reg_val_2 << 3);

    #else
    bm_print(BM_LOG_FULL,, "[Chip_Trim] need check reg number\n");
    #endif

    switch(chip_diff_trim_value_4_0){       
        case 0:    
            chip_diff_trim_value = 1000; 
            break;
        case 1:    chip_diff_trim_value = 1005; break;
        case 2:    chip_diff_trim_value = 1010; break;
        case 3:    chip_diff_trim_value = 1015; break;
        case 4:    chip_diff_trim_value = 1020; break;
        case 5:    chip_diff_trim_value = 1025; break;
        case 6:    chip_diff_trim_value = 1030; break;
        case 7:    chip_diff_trim_value = 1035; break;
        case 8:    chip_diff_trim_value = 1040; break;
        case 9:    chip_diff_trim_value = 1045; break;
        case 10:   chip_diff_trim_value = 1050; break;
        case 11:   chip_diff_trim_value = 1055; break;
        case 12:   chip_diff_trim_value = 1060; break;
        case 13:   chip_diff_trim_value = 1065; break;
        case 14:   chip_diff_trim_value = 1070; break;
        case 15:   chip_diff_trim_value = 1075; break;
        case 31:   chip_diff_trim_value = 995; break; 
        case 30:   chip_diff_trim_value = 990; break; 
        case 29:   chip_diff_trim_value = 985; break; 
        case 28:   chip_diff_trim_value = 980; break; 
        case 27:   chip_diff_trim_value = 975; break; 
        case 26:   chip_diff_trim_value = 970; break; 
        case 25:   chip_diff_trim_value = 965; break; 
        case 24:   chip_diff_trim_value = 960; break; 
        case 23:   chip_diff_trim_value = 955; break; 
        case 22:   chip_diff_trim_value = 950; break; 
        case 21:   chip_diff_trim_value = 945; break; 
        case 20:   chip_diff_trim_value = 940; break; 
        case 19:   chip_diff_trim_value = 935; break; 
        case 18:   chip_diff_trim_value = 930; break; 
        case 17:   chip_diff_trim_value = 925; break; 
        default:
            bm_print(BM_LOG_FULL, "[Chip_Trim] Invalid value(%d)\n", chip_diff_trim_value_4_0);
            break;
    }

#endif
}

kal_int32 use_chip_trim_value(kal_int32 not_trim_val)
{
#if defined(CONFIG_POWER_EXT)
    return not_trim_val;
#else
    kal_int32 ret_val=0;

    ret_val=((not_trim_val*chip_diff_trim_value)/1000);

    return ret_val;
#endif    
}


int get_hw_ocv(void)
{
#if defined(CONFIG_POWER_EXT)
    return 4001;    
#else
    kal_int32 adc_result_reg=0;
    kal_int32 adc_result=0;
    kal_int32 r_val_temp=4;    

    #if defined(SWCHR_POWER_PATH)
    adc_result_reg = pmic_get_register_value(PMIC_RG_ADC_OUT_WAKEUP_SWCHR);
    adc_result = (adc_result_reg*r_val_temp*VOLTAGE_FULL_RANGE)/ADC_PRECISE;
    #else
    adc_result_reg = pmic_get_register_value(PMIC_RG_ADC_OUT_WAKEUP_PCHR);
    adc_result = (adc_result_reg*r_val_temp*VOLTAGE_FULL_RANGE)/ADC_PRECISE;        
    #endif

    adc_result += g_hw_ocv_tune_value;
    return adc_result;
#endif    
}


//============================================================//

static kal_int32 fgauge_initialization(void *data)
{
    return STATUS_OK;
}

static kal_int32 fgauge_read_current(void *data)
{
    return STATUS_OK;
}

static kal_int32 fgauge_read_current_sign(void *data)
{
    return STATUS_OK;
}

static kal_int32 fgauge_read_columb(void *data)
{
    return STATUS_OK;
}


static kal_int32 fgauge_hw_reset(void *data)
{
    return STATUS_OK;
}


static kal_int32 read_adc_v_bat_sense(void *data)
{
#if defined(CONFIG_POWER_EXT)
    *(kal_int32*)(data) = 4201;
#else
    *(kal_int32*)(data) = PMIC_IMM_GetOneChannelValue(MT6350_AUX_BATSNS,*(kal_int32*)(data),1);
#endif

    return STATUS_OK;
}



static kal_int32 read_adc_v_i_sense(void *data)
{
#if defined(CONFIG_POWER_EXT)
    *(kal_int32*)(data) = 4202;
#else
    *(kal_int32*)(data) = PMIC_IMM_GetOneChannelValue(MT6350_AUX_ISENSE,*(kal_int32*)(data),1);
#endif

    return STATUS_OK;
}

static kal_int32 read_adc_v_bat_temp(void *data)
{
#if defined(CONFIG_POWER_EXT)
    *(kal_int32*)(data) = 0;
#else
    #if defined(MTK_PCB_TBAT_FEATURE)

        int ret = 0, data[4], i, ret_value = 0, ret_temp = 0;
        int Channel=1;
        
        if( IMM_IsAdcInitReady() == 0 )
        {
            return 0;
        }
    
        i = times;
        while (i--)
        {
            ret_value = IMM_GetOneChannelValue(Channel, data, &ret_temp);
            ret += ret_temp;
        }
        
        ret = ret*1500/4096 ;
        ret = ret/times;

        *(kal_int32*)(data) = ret;

    #else
        *(kal_int32*)(data) = PMIC_IMM_GetOneChannelValue(MT6350_AUX_BATON1,*(kal_int32*)(data),1);
    #endif
#endif

    return STATUS_OK;
}

static kal_int32 read_adc_v_charger(void *data)
{
	kal_int32 val;
#if defined(CONFIG_POWER_EXT)
    *(kal_int32*)(data) = 5001;
#else
    val = PMIC_IMM_GetOneChannelValue(MT6350_AUX_VCDT,*(kal_int32*)(data),1);
	val = (((R_CHARGER_1+R_CHARGER_2)*100*val)/R_CHARGER_2)/100;
	
	*(kal_int32*)(data) = val;
#endif

    return STATUS_OK;
}

static kal_int32 read_hw_ocv(void *data)
{
#if defined(CONFIG_POWER_EXT)
    *(kal_int32*)(data) = 3999;
#else
    *(kal_int32*)(data) = get_hw_ocv();
#endif

    return STATUS_OK;
}

static kal_int32 dump_register_fgadc(void *data)
{
    return STATUS_OK;
}

static kal_int32 read_battery_plug_out_status(void *data)
{
	*(kal_int32*)(data) = get_battery_plug_out_status();

	return STATUS_OK;
}

static kal_int32 fgauge_set_columb_interrupt(void *data)
{
    return STATUS_OK;
}

static kal_int32 fgauge_read_columb_accurate(void *data)
{
    return 0;
}


#if defined(CONFIG_MTK_HAFG_20)
kal_int8 is_lbat_int=KAL_TRUE;
void fg_bat_l_int_handler(void)
{
	pmic_set_register_value(PMIC_RG_LBAT_IRQ_EN_MIN,0);
	pmic_set_register_value(PMIC_RG_LBAT_EN_MIN,0);
	pmic_set_register_value(PMIC_RG_INT_EN_BAT_L,0);
	is_lbat_int=KAL_TRUE;

}


extern void pmic_register_interrupt_callback(kal_uint32 intNo,void (EINT_FUNC_PTR)(void));
extern void pmic_turn_on_clock(kal_uint32 enable);
#endif
static kal_int32 fgauge_set_low_battery_interrupt(void *data)
{
#if defined(CONFIG_MTK_HAFG_20)
	kal_uint32 voltage=*(kal_uint32*)(data);
	kal_uint32 rawdata;
	static int eint_init=0;

	if(eint_init==0)
	{
		pmic_register_interrupt_callback(2,fg_bat_l_int_handler);
		eint_init=1;
	}

	is_lbat_int=KAL_FALSE;

	pmic_turn_on_clock(KAL_TRUE);

	rawdata=voltage*4096/1800/4;

	// 0:set interrupt 
	pmic_set_register_value(PMIC_RG_INT_EN_BAT_L,1);

	// 1.setup min voltage threshold 3.4v
	pmic_set_register_value(PMIC_RG_LBAT_VOLT_MIN,rawdata);	

	// 2.setup detection period
	pmic_set_register_value(PMIC_RG_LBAT_DET_PRD_19_16,0x0);
	pmic_set_register_value(PMIC_RG_LBAT_DET_PRD_15_0,0x12f);

	// 3.setup max/min debounce time
	pmic_set_register_value(PMIC_RG_LBAT_DEBT_MIN,0x5a);//15mins	

	// 4.turn on IRQ
	pmic_set_register_value(PMIC_RG_LBAT_IRQ_EN_MIN,0x1); 	

	// 5. turn on LowBattery Detection
	pmic_set_register_value(PMIC_RG_LBAT_EN_MIN,1);
#endif	

	return STATUS_OK;
}

static kal_int32 fgauge_get_low_battery_interrupt_status(void *data)
{
#if defined(CONFIG_MTK_HAFG_20)
	*(kal_int32*)(data) = is_lbat_int;
#else
	*(kal_int32*)(data) = KAL_FALSE;
#endif

		return STATUS_OK;
}


void trigger_hw_ocv(void)
{
	pmic_set_register_value(PMIC_STRUP_AUXADC_START_SEL,0x1); 	
	//udelay(50);
	pmic_set_register_value(PMIC_STRUP_AUXADC_START_SW,0x0);
	//udelay(50);
	pmic_set_register_value(PMIC_STRUP_AUXADC_START_SW,0x1); 

	while(pmic_get_register_value(PMIC_RG_ADC_RDY_WAKEUP_PCHR)==0)
	{
		udelay(100);
	}
}



extern void pmic_turn_on_clock(kal_uint32 enable);

static kal_int32 get_refresh_hw_ocv(void *data)
{
#if defined(CONFIG_MTK_HAFG_20)

		kal_int32 hwocv1,hwocv2;
		pmic_turn_on_clock(1);
		udelay(30);
		hwocv1=get_hw_ocv();
		trigger_hw_ocv();
		hwocv2=get_hw_ocv();
		pmic_set_register_value(PMIC_STRUP_AUXADC_START_SEL,0x0); 
		pmic_turn_on_clock(0);

		*(kal_int32*)(data)=hwocv2;
		
#else
		*(kal_int32*)(data)=0;
#endif
		return STATUS_OK;
}



static kal_int32 (* const bm_func[BATTERY_METER_CMD_NUMBER])(void *data)=
{
    fgauge_initialization		//hw fuel gague used only
    
    ,fgauge_read_current		//hw fuel gague used only
    ,fgauge_read_current_sign	//hw fuel gague used only
    ,fgauge_read_columb			//hw fuel gague used only

    ,fgauge_hw_reset			//hw fuel gague used only
    
    ,read_adc_v_bat_sense
    ,read_adc_v_i_sense
    ,read_adc_v_bat_temp
    ,read_adc_v_charger

    ,read_hw_ocv
    ,dump_register_fgadc		//hw fuel gague used only
    ,fgauge_set_columb_interrupt 
    ,read_battery_plug_out_status    
    ,fgauge_read_columb_accurate    
    ,fgauge_set_low_battery_interrupt
    ,fgauge_get_low_battery_interrupt_status
    ,get_refresh_hw_ocv
};

kal_int32 bm_ctrl_cmd(BATTERY_METER_CTRL_CMD cmd, void *data)
{
    kal_int32 status;

    if(cmd < BATTERY_METER_CMD_NUMBER)
    {
    	if(bm_func[cmd]==NULL)
    	{
    	}
		else
		{
        	status = bm_func[cmd](data);
		}
    }
    else
        return STATUS_UNSUPPORTED;
    
    return status;
}

