/*****************************************************************************
 *
 * Filename:
 * ---------
 *    pmic.c
 *
 * Project:
 * --------
 *   Android_Software
 *
 * Description:
 * ------------
 *   This Module defines PMIC functions
 *
 * Author:
 * -------
 * WY Chuang
 *
 ****************************************************************************/
#include <generated/autoconf.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>
#include <linux/list.h>
#include <linux/mutex.h>
#include <linux/kthread.h>
#include <linux/wakelock.h>
#include <linux/device.h>
#include <linux/kdev_t.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/delay.h>
#include <linux/platform_device.h>
#include <linux/aee.h>
#include <linux/xlog.h>
#include <linux/proc_fs.h>
#include <linux/syscalls.h>
#include <linux/sched.h>
#include <linux/writeback.h>
#include <linux/earlysuspend.h>
#include <linux/seq_file.h>

#include <asm/uaccess.h>

#include <linux/regulator/driver.h>   
#include <linux/regulator/machine.h>   


#include <mach/upmu_common.h>
#include <mach/upmu_sw.h>
#include <mach/upmu_hw.h>
#include <mach/pmic.h>
#include <mach/mt_pm_ldo.h>
#include <mach/eint.h>
#include <mach/mt_pmic_wrap.h>
#include <mach/mt_gpio.h>
#include <mach/mtk_rtc.h>
#include <mach/mt_spm_mtcmos.h>

#include <mach/battery_common.h>
#include <linux/time.h>

#include <mach/mt_pm_ldo.h>
#include "pmic_dvt.h"

#if defined (CONFIG_MTK_KERNEL_POWER_OFF_CHARGING)
#include <mach/mt_boot.h>
#include <mach/system.h>
#include "mach/mt_gpt.h"
#endif

#include <mach/battery_meter.h>
#include <cust_pmic.h>
#include <cust_eint.h>
#include <cust_battery_meter.h>
#include <mtk_kpd.h>

 
//==============================================================================
// PMIC extern functions
//==============================================================================
extern kal_uint16 pmic_set_register_value(PMU_FLAGS_LIST_ENUM flagname,kal_uint32 val);
extern kal_uint16 pmic_get_register_value(PMU_FLAGS_LIST_ENUM flagname);
extern void pmic_auxadc_init(void);

//==============================================================================
// PMIC extern variable
//==============================================================================
#if defined (CONFIG_MTK_KERNEL_POWER_OFF_CHARGING)
extern void mt_power_off(void);
static kal_bool long_pwrkey_press = false;
static unsigned long timer_pre = 0; 
static unsigned long timer_pos = 0; 
#define LONG_PWRKEY_PRESS_TIME         500*1000000    //500ms
#endif

//==============================================================================
// Global variable
//==============================================================================
unsigned int g_eint_pmic_num = 17;
unsigned int g_cust_eint_mt_pmic_debounce_cn = 1; 
unsigned int g_cust_eint_mt_pmic_type = 4;
unsigned int g_cust_eint_mt_pmic_debounce_en = 1;

//==============================================================================
// PMIC related define
//==============================================================================
static DEFINE_MUTEX(pmic_lock_mutex);
#define PMIC_EINT_SERVICE

//==============================================================================
// PMIC read/write APIs
//==============================================================================
#if 0 //defined(CONFIG_MTK_FPGA)
    // no CONFIG_PMIC_HW_ACCESS_EN
#else
    #define CONFIG_PMIC_HW_ACCESS_EN
#endif

#define PMICLOG(fmt, arg...)   printk(PMICTAG fmt,##arg)
//#define PMICLOG(fmt, arg...) do{}while(0)


static DEFINE_MUTEX(pmic_access_mutex);

U32 pmic_read_interface (U32 RegNum, U32 *val, U32 MASK, U32 SHIFT)
{
    U32 return_value = 0;

#if defined(CONFIG_PMIC_HW_ACCESS_EN)
    U32 pmic_reg = 0;
    U32 rdata;

    mutex_lock(&pmic_access_mutex);

    //mt_read_byte(RegNum, &pmic_reg);
    return_value= pwrap_wacs2(0, (RegNum), 0, &rdata);
    pmic_reg=rdata;
    if(return_value!=0)
    {
        PMICLOG("[pmic_read_interface] Reg[%x]= pmic_wrap read data fail\n", RegNum);
        mutex_unlock(&pmic_access_mutex);
        return return_value;
    }
    //PMICLOG"[pmic_read_interface] Reg[%x]=0x%x\n", RegNum, pmic_reg);

    pmic_reg &= (MASK << SHIFT);
    *val = (pmic_reg >> SHIFT);
    //PMICLOG"[pmic_read_interface] val=0x%x\n", *val);

    mutex_unlock(&pmic_access_mutex);
#else
    //PMICLOG("[pmic_read_interface] Can not access HW PMIC\n");
#endif

    return return_value;
}

U32 pmic_config_interface (U32 RegNum, U32 val, U32 MASK, U32 SHIFT)
{
    U32 return_value = 0;

#if defined(CONFIG_PMIC_HW_ACCESS_EN)
    U32 pmic_reg = 0;
    U32 rdata;

    mutex_lock(&pmic_access_mutex);

    //1. mt_read_byte(RegNum, &pmic_reg);
    return_value= pwrap_wacs2(0, (RegNum), 0, &rdata);
    pmic_reg=rdata;
    if(return_value!=0)
    {
        PMICLOG("[pmic_config_interface] Reg[%x]= pmic_wrap read data fail\n", RegNum);
        mutex_unlock(&pmic_access_mutex);
        return return_value;
    }
    //PMICLOG"[pmic_config_interface] Reg[%x]=0x%x\n", RegNum, pmic_reg);

    pmic_reg &= ~(MASK << SHIFT);
    pmic_reg |= (val << SHIFT);

    //2. mt_write_byte(RegNum, pmic_reg);
    return_value= pwrap_wacs2(1, (RegNum), pmic_reg, &rdata);
    if(return_value!=0)
    {
        PMICLOG("[pmic_config_interface] Reg[%x]= pmic_wrap read data fail\n", RegNum);
        mutex_unlock(&pmic_access_mutex);
        return return_value;
    }
    //PMICLOG"[pmic_config_interface] write Reg[%x]=0x%x\n", RegNum, pmic_reg);

    #if 0
    //3. Double Check
    //mt_read_byte(RegNum, &pmic_reg);
    return_value= pwrap_wacs2(0, (RegNum), 0, &rdata);
    pmic_reg=rdata;
    if(return_value!=0)
    {
        PMICLOG("[pmic_config_interface] Reg[%x]= pmic_wrap write data fail\n", RegNum);
        mutex_unlock(&pmic_access_mutex);
        return return_value;
    }
    PMICLOG("[pmic_config_interface] Reg[%x]=0x%x\n", RegNum, pmic_reg);
    #endif

    mutex_unlock(&pmic_access_mutex);
#else
    //PMICLOG("[pmic_config_interface] Can not access HW PMIC\n");
#endif

    return return_value;
}


//==============================================================================
// PMIC lock/unlock APIs
//==============================================================================
void pmic_lock(void)
{
    mutex_lock(&pmic_lock_mutex);
}

void pmic_unlock(void)
{
    mutex_unlock(&pmic_lock_mutex);
}

kal_uint32 upmu_get_reg_value(kal_uint32 reg)
{
    U32 ret=0;
    U32 reg_val=0;
    
    ret=pmic_read_interface(reg, &reg_val, 0xFFFF, 0x0);
    
    return reg_val;
}
EXPORT_SYMBOL(upmu_get_reg_value);

void upmu_set_reg_value(kal_uint32 reg, kal_uint32 reg_val)
{
    U32 ret=0;
    
    ret=pmic_config_interface(reg, reg_val, 0xFFFF, 0x0);    
}

/**********************************************************
  *
  *   [Internal Function] 
  *
  *********************************************************/
void mt6350_dump_register(void)
{
    kal_uint8 i=0;

    PMICLOG("dump PMIC 6350 register\n");

   for(i=0;i<=0x079a;i=i+10)
   {
    PMICLOG("Reg[0x%x]=0x%x Reg[0x%x]=0x%x Reg[0x%x]=0x%x Reg[0x%x]=0x%x Reg[0x%x]=0x%x\n",
        i,upmu_get_reg_value(i),i+1,upmu_get_reg_value(i+1),i+2,upmu_get_reg_value(i+2),i+3,upmu_get_reg_value(i+3),i+4,upmu_get_reg_value(i+4));

       PMICLOG("Reg[0x%x]=0x%x Reg[0x%x]=0x%x Reg[0x%x]=0x%x Reg[0x%x]=0x%x Reg[0x%x]=0x%x\n",
            i+5,upmu_get_reg_value(i+5),i+6,upmu_get_reg_value(i+6),i+7,upmu_get_reg_value(i+7),i+8,upmu_get_reg_value(i+8),i+9,upmu_get_reg_value(i+9));
   }

}



//==============================================================================
// upmu_interrupt_chrdet_int_en
//==============================================================================
void upmu_interrupt_chrdet_int_en(kal_uint32 val)
{
    pmic_set_register_value(PMIC_RG_INT_EN_CHRDET,val);
}
EXPORT_SYMBOL(upmu_interrupt_chrdet_int_en);


//==============================================================================
// PMIC charger detection
//==============================================================================
kal_uint32 upmu_get_rgs_chrdet(void)
{
    kal_uint32 val=0;


    val = pmic_get_register_value(PMIC_RGS_CHRDET);

    return val;
}

//==============================================================================
// mt-pmic dev_attr APIs
//==============================================================================
U32 g_reg_value=0;
static ssize_t show_pmic_access(struct device *dev,struct device_attribute *attr, char *buf)
{
    return sprintf(buf, "%u\n", g_reg_value);
}
static ssize_t store_pmic_access(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
    int ret=0;
    char *pvalue = NULL;
    U32 reg_value = 0;
    U32 reg_address = 0;
    if(buf != NULL && size != 0)
    {
        reg_address = simple_strtoul(buf,&pvalue,16);

        if(size > 5)
        {
            reg_value = simple_strtoul((pvalue+1),NULL,16);
            ret=pmic_config_interface(reg_address, reg_value, 0xFFFF, 0x0);
        }
        else
        {
            ret=pmic_read_interface(reg_address, &g_reg_value, 0xFFFF, 0x0);
        }
    }
    return size;
}
static DEVICE_ATTR(pmic_access, 0664, show_pmic_access, store_pmic_access); //664

//==============================================================================
// DVT entry
//==============================================================================
kal_uint8 g_reg_value_pmic=0;

static ssize_t show_pmic_dvt(struct device *dev,struct device_attribute *attr, char *buf)
{
    return sprintf(buf, "%u\n", g_reg_value_pmic);
}

static ssize_t store_pmic_dvt(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
    char *pvalue = NULL;
    unsigned int test_item = 0;
    
    if(buf != NULL && size != 0)
    {
        test_item = simple_strtoul(buf,&pvalue,10);

        #ifdef MTK_PMIC_DVT_SUPPORT 
        pmic_dvt_entry(test_item);
        #endif
    }    
    return size;
}
static DEVICE_ATTR(pmic_dvt, 0664, show_pmic_dvt, store_pmic_dvt);


//==============================================================================
// PMIC6328 linux reguplator driver
//==============================================================================
extern PMU_FLAG_TABLE_ENTRY pmu_flags_table[];

static int mtk_regulator_enable(struct regulator_dev *rdev)
{
    const struct regulator_desc *rdesc=rdev->desc;
    struct mtk_regulator *mreg;
    kal_uint32 add=0,val=0;

    mreg=container_of(rdesc,struct mtk_regulator,desc);
    if (mreg->en_reg!=0)
    {
        pmic_set_register_value(mreg->en_reg,1);
        add=pmu_flags_table[mreg->en_reg].offset;
        val=upmu_get_reg_value(pmu_flags_table[mreg->en_reg].offset);
    }

    return 0;
}

static int mtk_regulator_disable(struct regulator_dev *rdev)
{
    const struct regulator_desc *rdesc=rdev->desc;
    struct mtk_regulator *mreg;
    kal_uint32 add=0,val=0;

    mreg=container_of(rdesc,struct mtk_regulator,desc);
        
    if (rdev->use_count==0)
    {
        PMICLOG("regulator_disable fail (name=%s use_count=%d) \n",rdesc->name,rdev->use_count);
        return -1;
    }

    if (mreg->en_reg!=0)
    {
        pmic_set_register_value(mreg->en_reg,0);
        add=pmu_flags_table[mreg->en_reg].offset;
        val=upmu_get_reg_value(pmu_flags_table[mreg->en_reg].offset);       
    }

    return 0;
}

static int mtk_regulator_is_enabled(struct regulator_dev *rdev)
{
    const struct regulator_desc *rdesc=rdev->desc;
    struct mtk_regulator *mreg;
    int en;

    mreg=container_of(rdesc,struct mtk_regulator,desc);

    en=pmic_get_register_value(mreg->en_reg);
        
    return en;
}

static int mtk_regulator_get_voltage_sel(struct regulator_dev *rdev)
{
    const struct regulator_desc *rdesc=rdev->desc;
    struct mtk_regulator *mreg;
    unsigned char regVal=0;
    const int *pVoltage;
    int voltage=0;
    kal_uint32 add=0,val=9;

    mreg=container_of(rdesc,struct mtk_regulator,desc);
    
    if(mreg->desc.n_voltages!=1)
    {
        if(mreg->vol_reg!=0)
        {
            regVal=pmic_get_register_value(mreg->vol_reg);
            if(mreg->pvoltages!=NULL)
            {
                pVoltage=(const int*)mreg->pvoltages;
                voltage=pVoltage[regVal];
                add=pmu_flags_table[mreg->en_reg].offset;
                val=upmu_get_reg_value(pmu_flags_table[mreg->en_reg].offset);
            }
            else
            {
                voltage=mreg->desc.min_uV+mreg->desc.uV_step*regVal;
            }
        }
        else
        {
        }
    }
    else
    {
        if(mreg->vol_reg!=0)
        {
            regVal=0;
            pVoltage=(const int*)mreg->pvoltages;
            voltage=pVoltage[regVal];
        }
        else
        {
        }
    }
    
    return regVal;
}

static int mtk_regulator_set_voltage_sel(struct regulator_dev *rdev,
                          unsigned selector)
{
    const struct regulator_desc *rdesc=rdev->desc;
    struct mtk_regulator *mreg;

    mreg=container_of(rdesc,struct mtk_regulator,desc);
        
/* VGP2
    0:1200000,->0
    1:1300000,->1
    2:1500000,->2
    3:1800000,->3
    4:2000000,->7
    5:2500000,->4
    6:2800000,->5
    7:3000000,->6 
*/	
        if(strcmp(rdesc->name, "VGP2")==0)
        {
                if (mreg->vol_reg != 0)
                {       
                        if(selector <= 3)
                        {
                                pmic_set_register_value(mreg->vol_reg, selector);
                        }
                        else if(selector == 4)
                        {
                                pmic_set_register_value(mreg->vol_reg, selector + 3);
                        }
                        else
                        {
                                pmic_set_register_value(mreg->vol_reg, selector - 1);
                        }
                }				
        }    
        else
        {
    if (mreg->vol_reg!=0)
    {
        pmic_set_register_value(mreg->vol_reg,selector);
    }
        }

    return 0;
}

static int mtk_regulator_list_voltage(struct regulator_dev *rdev,
                       unsigned selector)
{
    const struct regulator_desc *rdesc=rdev->desc;
    struct mtk_regulator *mreg;
    const int *pVoltage;
    int voltage=0;

    mreg=container_of(rdesc,struct mtk_regulator,desc);
        
    if(mreg->desc.n_voltages!=1)
    {
        if(mreg->vol_reg!=0)
        {
            if(mreg->pvoltages!=NULL)
            {
                pVoltage=(const int*)mreg->pvoltages;
                voltage=pVoltage[selector];
            }
            else
            {
                voltage=mreg->desc.min_uV+mreg->desc.uV_step*selector;
            }
        }
        else
        {
        }
    }
    else
    {
                pVoltage=(const int*)mreg->pvoltages;
                voltage=pVoltage[0];
    }
    
//  PMICLOG("regulator_list_voltage(name=%s id=%d en_reg=%x vol_reg=%x selector=%d voltage=%d)\n",rdesc->name, rdesc->id
//      ,mreg->en_reg,mreg->vol_reg,selector,voltage);

    return voltage;
}




static struct regulator_ops mtk_regulator_ops = {
    .enable          = mtk_regulator_enable,
    .disable         = mtk_regulator_disable,
    .is_enabled      = mtk_regulator_is_enabled,
    .get_voltage_sel = mtk_regulator_get_voltage_sel,
    .set_voltage_sel = mtk_regulator_set_voltage_sel,
    .list_voltage    = mtk_regulator_list_voltage,
};

static ssize_t show_LDO_STATUS(struct device *dev,struct device_attribute *attr, char *buf)
{
    struct mtk_regulator *mreg;
    unsigned int ret_value=0;

    mreg=container_of(attr,struct mtk_regulator,en_att);

    ret_value=pmic_get_register_value(mreg->en_reg);

    return sprintf(buf, "%u\n", ret_value); 
}

static ssize_t store_LDO_STATUS(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
    return size;
}

static ssize_t show_LDO_VOLTAGE(struct device *dev,struct device_attribute *attr, char *buf)
{
    struct mtk_regulator *mreg;
    const int *pVoltage;

    kal_uint16 regVal;
    unsigned int ret_value=0;

    mreg=container_of(attr,struct mtk_regulator,voltage_att);

    if(mreg->desc.n_voltages!=1)
    {
        if(mreg->vol_reg!=0)
        {
            regVal=pmic_get_register_value(mreg->vol_reg);
            if(mreg->pvoltages!=NULL)
            {
                pVoltage=(const int*)mreg->pvoltages;
                ret_value=pVoltage[regVal];
            }
            else
            {
                ret_value=mreg->desc.min_uV+mreg->desc.uV_step*regVal;
            }
        }
        else
        {
        }
    }
    else
    {
        pVoltage=(const int*)mreg->pvoltages;
        ret_value=pVoltage[0];
    }

    ret_value=ret_value/1000;
    return sprintf(buf, "%u\n", ret_value);

}

static ssize_t store_LDO_VOLTAGE(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
    return size;
}

static ssize_t show_BUCK_STATUS(struct device *dev,struct device_attribute *attr, char *buf)
{
    struct mtk_regulator *mreg;
    unsigned int ret_value=0;

    mreg=container_of(attr,struct mtk_regulator,en_att);

    ret_value=pmic_get_register_value(mreg->qi_en_reg);

    return sprintf(buf, "%u\n", ret_value); 
}
static ssize_t store_BUCK_STATUS(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
    return size;
}

static ssize_t show_BUCK_VOLTAGE(struct device *dev,struct device_attribute *attr, char *buf)
{
    struct mtk_regulator *mreg;
    const int *pVoltage;

    kal_uint16 regVal;
    unsigned int ret_value=0;

    mreg=container_of(attr,struct mtk_regulator,voltage_att);

    if(mreg->desc.n_voltages!=1)
    {
        if(mreg->qi_vol_reg!=0)
        {
            regVal=pmic_get_register_value(mreg->qi_vol_reg);
            if(mreg->pvoltages!=NULL)
            {
                pVoltage=(const int*)mreg->pvoltages;
                ret_value=pVoltage[regVal];
            }
            else
            {
                ret_value=mreg->desc.min_uV+mreg->desc.uV_step*regVal;
            }
        }
        else
        {
        }
    }
    else
    {
        pVoltage=(const int*)mreg->pvoltages;   
        ret_value=pVoltage[0];  
    }

    ret_value=ret_value/1000;
    return sprintf(buf, "%u\n", ret_value);
}
static ssize_t store_BUCK_VOLTAGE(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
    return size;
}

// LDO Voltage Table
const static int pmic_1v8_voltages[] = {
    1800000,
};

const static int pmic_2v2_voltages[] = {
    2200000,
};

const static int pmic_2v8_voltages[] = {
    2800000,
};

const static int pmic_3v3_voltages[] = {
    3300000,
};

const static int VA_voltages[] = {
    1800000,
    2500000,
};

const static int VCAMA_voltages[] = {
    1500000,
    1800000,
    2500000,
    2800000,    
};

const static int VCN33_voltages[] = {
    3300000,
    3400000,
    3500000,
    3600000,
};

const static int VCN28_voltages[] = {
    1800000,
    2800000,
};

const static int VCAMD_voltages[] = {
    1200000,
    1300000,    
    1500000,
    1800000,        
};

const static int VMC_voltages[] = {
    1800000,
    3300000,
};

const static int VMCH_voltages[] = {
    3000000,
    3300000,
};

const static int VEMC3V3_voltages[] = {
    3000000,
    3300000,
};

const static int VGP1_voltages[] = {
    1200000,
    1300000,
    1500000,
    1800000,
    2000000,
    2800000,
    3000000,
    3300000,    
};

const static int VGP2_voltages[] = {
    1200000,
    1300000,
    1500000,
    1800000,
    2000000,
    2500000,
    2800000,
    3000000,
};

const static int VGP3_voltages[] = {
    1200000,
    1300000,
    1500000,
    1800000,
};

const static int VCAMAF_voltages[] = {
    1200000,
    1300000,
    1500000,
    1800000,
    2000000,
    2800000,
    3000000,
    3300000,
};

const static int VSIM1_voltages[] = {
    1800000,
    3000000,    
};

const static int VSIM2_voltages[] = {
    1800000,
    3000000,    
};

const static int VIBR_voltages[] = {
    1200000,
    1300000,    
    1500000,
    1800000,    
    2000000,
    2800000,    
    3000000,
    3300000,        
};

const static int VM_voltages[] = {
    1200000,
    1350000,    
    1500000,
    1800000,        
};

/*
ALDO        VCN28
ALDO        VTCXO
ALDO        VA
ALDO        VCAMA
DLDO        VCN_3V3
DLDO        VIO28
DLDO        VSIM1
DLDO        VSIM2
DLDO        VUSB
DLDO        VGP1
DLDO        VGP2
DLDO        VEMC_3V3
DLDO        VCAM_AF
DLDO        VMC
DLDO        VMCH
DLDO        VIBR
RTCLDO      VRTC
VSYS LDO    VM
VSYS LDO    VRF18
VSYS LDO    VIO18
VSYS LDO    VCAMD
VSYS LDO    VCAM_IO
VSYS LDO    VGP3
VSYS LDO    VCN18
*/

struct mtk_regulator mtk_ldos[] = {
        PMIC_LDO_GEN1(VCN28,        PMIC_RG_VCN28_EN,       NULL,                   pmic_2v8_voltages,      1,PMIC_EN), 
        PMIC_LDO_GEN1(VTCXO,        PMIC_RG_VTCXO_EN,       NULL,                   pmic_2v2_voltages,      0,PMIC_EN),
        PMIC_LDO_GEN1(VA,           PMIC_RG_VA_EN,          PMIC_RG_VA_VOSEL,       VA_voltages,            1,PMIC_EN_VOL),
        PMIC_LDO_GEN1(VCAMA,        PMIC_RG_VCAMA_EN,       PMIC_RG_VCAMA_VOSEL,    VCAMA_voltages,         1,PMIC_EN_VOL),
        PMIC_LDO_GEN1(VCN33_BT,     PMIC_RG_VCN33_EN_BT,    PMIC_RG_VCN33_VOSEL,    VCN33_voltages,         1,PMIC_EN_VOL),
        PMIC_LDO_GEN1(VCN33_WIFI,   PMIC_RG_VCN33_EN_WIFI,  PMIC_RG_VCN33_VOSEL,    VCN33_voltages,         1,PMIC_EN_VOL),     
        PMIC_LDO_GEN1(VIO28,        PMIC_VIO28_EN,          NULL,                   pmic_2v8_voltages,      0,PMIC_EN), 
        PMIC_LDO_GEN1(VSIM1,        PMIC_RG_VSIM1_EN,       PMIC_RG_VSIM1_VOSEL,    VSIM1_voltages,         1,PMIC_EN_VOL), 
        PMIC_LDO_GEN1(VSIM2,        PMIC_RG_VSIM2_EN,       PMIC_RG_VSIM2_VOSEL,    VSIM2_voltages,         1,PMIC_EN_VOL), 
        PMIC_LDO_GEN1(VUSB,         PMIC_RG_VUSB_EN,        NULL,                   pmic_3v3_voltages,      1,PMIC_EN),
        PMIC_LDO_GEN1(VGP1,         PMIC_RG_VGP1_EN,        PMIC_RG_VGP1_VOSEL,     VGP1_voltages,          1,PMIC_EN_VOL), 
        PMIC_LDO_GEN1(VGP2,         PMIC_RG_VGP2_EN,        PMIC_RG_VGP2_VOSEL,     VGP2_voltages,          1,PMIC_EN_VOL),
        PMIC_LDO_GEN1(VEMC_3V3,     PMIC_RG_VEMC_3V3_EN,    PMIC_RG_VEMC_3V3_VOSEL, VEMC3V3_voltages,       1,PMIC_EN_VOL),
        PMIC_LDO_GEN1(VCAMAF,       PMIC_RG_VCAM_AF_EN,     PMIC_RG_VCAM_AF_VOSEL,  VCAMAF_voltages,        1,PMIC_EN_VOL),     
        PMIC_LDO_GEN1(VMC,          PMIC_RG_VMC_EN,         PMIC_RG_VMC_VOSEL,      VMC_voltages,           1,PMIC_EN_VOL), 
        PMIC_LDO_GEN1(VMCH,         PMIC_RG_VMCH_EN,        PMIC_RG_VMCH_VOSEL,     VMCH_voltages,          1,PMIC_EN_VOL), 
        PMIC_LDO_GEN1(VIBR,         PMIC_RG_VIBR_EN,        PMIC_RG_VIBR_VOSEL,     VIBR_voltages,          1,PMIC_EN_VOL),
        PMIC_LDO_GEN1(VRTC,         PMIC_VRTC_EN,           NULL,                   pmic_2v8_voltages,      0,PMIC_EN),
        PMIC_LDO_GEN1(VM,           PMIC_RG_VM_EN,          PMIC_RG_VM_VOSEL,       VM_voltages,            1,PMIC_EN_VOL),
        PMIC_LDO_GEN1(VRF18,        PMIC_RG_VRF18_EN,       NULL,                   pmic_1v8_voltages,      1,PMIC_EN), 
        PMIC_LDO_GEN1(VIO18,        PMIC_RG_VIO18_EN,       NULL,                   pmic_1v8_voltages,      1,PMIC_EN),
        PMIC_LDO_GEN1(VCAMD,        PMIC_RG_VCAMD_EN,       PMIC_RG_VCAMD_VOSEL,    VCAMD_voltages,         1,PMIC_EN_VOL), 
        PMIC_LDO_GEN1(VCAMIO,       PMIC_RG_VCAM_IO_EN,     NULL,                   pmic_1v8_voltages,      1,PMIC_EN), 
        PMIC_LDO_GEN1(VGP3,         PMIC_RG_VGP3_EN,        PMIC_RG_VGP3_VOSEL,     VGP3_voltages,          1,PMIC_EN_VOL), 
        PMIC_LDO_GEN1(VCN_1V8,      PMIC_RG_VCN_1V8_EN,     NULL              ,     pmic_1v8_voltages,      1,PMIC_EN), 
};

// Buck only provide query voltage & enable
static struct mtk_regulator mtk_bucks[] = {
        PMIC_BUCK_GEN(VPA,          PMIC_QI_VPA_EN,         PMIC_NI_VPA_VOSEL,      500000, 3650000,50000),
        PMIC_BUCK_GEN(VPROC,        PMIC_QI_VPROC_EN,       PMIC_NI_VPROC_VOSEL,    700000, 1493750,6250),
};

void mtk_regulator_init(struct platform_device *dev)
{
    int i=0;
    int ret=0;
    int isEn=0;

    for (i = 0; i < ARRAY_SIZE(mtk_ldos); i++) 
    {
        if (mtk_ldos[i].isUsedable==1)
        {
            mtk_ldos[i].config.dev=&(dev->dev);
            mtk_ldos[i].config.init_data=&mtk_ldos[i].init_data;
            if(mtk_ldos[i].desc.n_voltages!=1)
            {
                const int *pVoltage;
                
                if(mtk_ldos[i].vol_reg!=0)
                {
                    if(mtk_ldos[i].pvoltages!=NULL)
                    {
                        pVoltage=(const int*)mtk_ldos[i].pvoltages;
                        mtk_ldos[i].init_data.constraints.max_uV=pVoltage[mtk_ldos[i].desc.n_voltages-1];
                        mtk_ldos[i].init_data.constraints.min_uV=pVoltage[0];       
                    }
                    else
                    {
                        mtk_ldos[i].init_data.constraints.max_uV=(mtk_ldos[i].desc.n_voltages-1)*mtk_ldos[i].desc.uV_step+mtk_ldos[i].desc.min_uV;
                        mtk_ldos[i].init_data.constraints.min_uV=mtk_ldos[i].desc.min_uV;
                    }
                }
            }

            mtk_ldos[i].desc.owner=THIS_MODULE;

            mtk_ldos[i].rdev= regulator_register(&mtk_ldos[i].desc,&mtk_ldos[i].config);
            if ( IS_ERR(mtk_ldos[i].rdev) ) {                                     
                ret = PTR_ERR(mtk_ldos[i].rdev);
                PMICLOG("[regulator_register] failed to register %s (%d)\n", mtk_ldos[i].desc.name, ret);
            } else {
            }
            mtk_ldos[i].reg=regulator_get(&(dev->dev), mtk_ldos[i].desc.name);
            isEn=regulator_is_enabled(mtk_ldos[i].reg); 
            if(isEn!=0)
            {
                //ret=regulator_enable(mtk_ldos[i].reg);
            }           
        }
    }

}



void pmic6350_regulator_test(void)
{
    int i=0,j;
    int ret1,ret2;
    struct regulator *reg;
    
    for (i = 0; i < ARRAY_SIZE(mtk_ldos); i++) 
    {
        if (mtk_ldos[i].isUsedable==1)
        {
            reg=mtk_ldos[i].reg;
            
            ret1=regulator_enable(reg);
            ret2=regulator_is_enabled(reg);

            if(ret2==pmic_get_register_value(mtk_ldos[i].en_reg))
            {
            }
            else
            {
                PMICLOG("[enable test fail] ret=%d enable=%d addr:0x%x reg:0x%x\n", ret1,ret2,
                pmu_flags_table[mtk_ldos[i].en_reg].offset,
                pmic_get_register_value(mtk_ldos[i].en_reg));
            }

            ret1=regulator_disable(reg);
            ret2=regulator_is_enabled(reg);

            if(ret2==pmic_get_register_value(mtk_ldos[i].en_reg))
            {
            }
            else
            {
                PMICLOG("[disable test fail] ret=%d enable=%d addr:0x%x reg:0x%x\n", ret1,ret2,
                pmu_flags_table[mtk_ldos[i].en_reg].offset,
                pmic_get_register_value(mtk_ldos[i].en_reg));
            }

        }
    }

    for (i = 0; i < ARRAY_SIZE(mtk_ldos); i++) 
    {
        const int *pVoltage;
    
        reg=mtk_ldos[i].reg;
        if (mtk_ldos[i].isUsedable==1)
        {
            if(mtk_ldos[i].pvoltages!=NULL)
            {
                pVoltage=(const int*)mtk_ldos[i].pvoltages;
        
                for(j=0;j<mtk_ldos[i].desc.n_voltages;j++)
                {
                    int rvoltage;
                    regulator_set_voltage(reg, pVoltage[j], pVoltage[j]);
                    rvoltage=regulator_get_voltage(reg);


                    if((j==pmic_get_register_value(mtk_ldos[i].vol_reg))&& (pVoltage[j]==rvoltage))
                    {
                    }
                    else
                    {
                        PMICLOG("[%d:%d]:fail  set_voltage:%d  rvoltage:%d \n",j,pmic_get_register_value(mtk_ldos[i].vol_reg),
                        pVoltage[j] ,rvoltage);                     
                    }       
                }
            }
        }
    }


}
/*
int getHwVoltage(MT65XX_POWER powerId)
{
    struct regulator *reg;
    
    
    if(powerId>=ARRAY_SIZE(mtk_ldos))
        return -100;

    if(mtk_ldos[powerId].isUsedable!=TRUE)
        return -101;
    
    reg=mtk_ldos[powerId].reg;

    return regulator_get_voltage(reg);
}
EXPORT_SYMBOL(getHwVoltage);

int isHwPowerOn(MT65XX_POWER powerId)
{
    struct regulator *reg;
    
    if(powerId>=ARRAY_SIZE(mtk_ldos))
        return -100;

    if(mtk_ldos[powerId].isUsedable!=TRUE)
        return -101;
    
    reg=mtk_ldos[powerId].reg;

    return regulator_is_enabled(reg);

}
EXPORT_SYMBOL(isHwPowerOn);

bool hwPowerOn(MT65XX_POWER powerId, int powerVolt, char *mode_name)
{
    struct regulator *reg;
    int ret1,ret2;
    
    if(powerId>=ARRAY_SIZE(mtk_ldos))
        return FALSE;

    if(mtk_ldos[powerId].isUsedable!=TRUE)
        return FALSE;
    
    reg=mtk_ldos[powerId].reg;

    ret2=regulator_set_voltage(reg, powerVolt, powerVolt);

    if (ret2!=0)
    {
        PMICLOG("hwPowerOn:%s set voltage %d fail",mtk_ldos[powerId].desc.name,powerVolt);
    }
    
    ret1=regulator_enable(reg);

    if (ret1!=0)
    {
        PMICLOG("hwPowerOn:%s enable fail",mtk_ldos[powerId].desc.name);
    }

    PMICLOG("hwPowerOn:%d:%s volt:%d name:%s cnt:%d",powerId,mtk_ldos[powerId].desc.name,powerVolt,mode_name,mtk_ldos[powerId].rdev->use_count);
    return TRUE;
}
EXPORT_SYMBOL(hwPowerOn);

bool hwPowerDown(MT65XX_POWER powerId, char *mode_name)
{
    struct regulator *reg;
    int ret1;
    
    if(powerId>=ARRAY_SIZE(mtk_ldos))
        return FALSE;
    
    reg=mtk_ldos[powerId].reg;
    ret1=regulator_disable(reg);

    if (ret1!=0)
    {
        PMICLOG("hwPowerOn err:ret1:%d ",ret1);
    }

    PMICLOG("hwPowerDown:%d:%s name:%s cnt:%d",powerId,mtk_ldos[powerId].desc.name,mode_name,mtk_ldos[powerId].rdev->use_count);
    return TRUE;
}
EXPORT_SYMBOL(hwPowerDown);

*/

//==============================================================================
// Low battery call back function
//==============================================================================
#define LBCB_NUM 16

#define DISABLE_LOW_BATTERY_PROTECT

#ifndef DISABLE_LOW_BATTERY_PROTECT
#define LOW_BATTERY_PROTECT
#endif


#ifdef LOW_BATTERY_PROTECT
// ex. 3400/5400*4096
#define BAT_HV_THD   (POWER_INT0_VOLT*4096/5400) //ex: 3400mV
#define BAT_LV_1_THD (POWER_INT1_VOLT*4096/5400) //ex: 3250mV
#define BAT_LV_2_THD (POWER_INT2_VOLT*4096/5400) //ex: 3000mV

int g_low_battery_level=0;
int g_low_battery_stop=0;

struct low_battery_callback_table
{
    void *lbcb;
};

struct low_battery_callback_table lbcb_tb[] ={
    {NULL},{NULL},{NULL},{NULL},{NULL},{NULL},{NULL},{NULL},
    {NULL},{NULL},{NULL},{NULL},{NULL},{NULL},{NULL},{NULL}    
};

void (*low_battery_callback)(LOW_BATTERY_LEVEL);

void register_low_battery_notify( void (*low_battery_callback)(LOW_BATTERY_LEVEL), LOW_BATTERY_PRIO prio_val )
{

    lbcb_tb[prio_val].lbcb = low_battery_callback;
    
}

void exec_low_battery_callback(LOW_BATTERY_LEVEL low_battery_level) //0:no limit
{
    int i=0;

    if(g_low_battery_stop==1)
    {
    }
    else
    {
        for(i=0 ; i<LBCB_NUM ; i++) 
        {
            if(lbcb_tb[i].lbcb != NULL)
            {
                low_battery_callback = lbcb_tb[i].lbcb;
                low_battery_callback(low_battery_level);
            }        
        }
    }
}

void lbat_min_en_setting(int en_val)
{
    pmic_set_register_value(PMIC_AUXADC_LBAT_EN_MIN,en_val);    
    pmic_set_register_value(PMIC_AUXADC_LBAT_IRQ_EN_MIN,en_val);    
    pmic_set_register_value(PMIC_RG_INT_EN_BAT_L,en_val);       
}

void lbat_max_en_setting(int en_val)
{
    pmic_set_register_value(PMIC_AUXADC_LBAT_EN_MAX,en_val);    
    pmic_set_register_value(PMIC_AUXADC_LBAT_IRQ_EN_MAX,en_val);    
    pmic_set_register_value(PMIC_RG_INT_EN_BAT_H,en_val);       
}

void low_battery_protect_init(void)
{
    //default setting
    pmic_set_register_value(PMIC_AUXADC_LBAT_DEBT_MIN,0);    
    pmic_set_register_value(PMIC_AUXADC_LBAT_DEBT_MAX,0);    
    pmic_set_register_value(PMIC_AUXADC_LBAT_DET_PRD_15_0,1);    
    pmic_set_register_value(PMIC_AUXADC_LBAT_DET_PRD_19_16,0);    

    pmic_set_register_value(PMIC_AUXADC_LBAT_VOLT_MAX,BAT_HV_THD);    
    pmic_set_register_value(PMIC_AUXADC_LBAT_VOLT_MIN,BAT_LV_1_THD);        

    lbat_min_en_setting(1);
    lbat_max_en_setting(0);
}

#endif //#ifdef LOW_BATTERY_PROTECT

void bat_h_int_handler(void)
{
    //sub-task
#ifdef LOW_BATTERY_PROTECT
    g_low_battery_level=0;
    exec_low_battery_callback(LOW_BATTERY_LEVEL_0);

#if 0
    lbat_max_en_setting(0);
    mdelay(1);
    lbat_min_en_setting(1);
#else
    pmic_set_register_value(PMIC_AUXADC_LBAT_VOLT_MIN,BAT_LV_1_THD); 

    lbat_min_en_setting(0);
    lbat_max_en_setting(0);
    mdelay(1);
    lbat_min_en_setting(1);   
#endif

#endif 

}

void bat_l_int_handler(void)
{
    //sub-task
#ifdef LOW_BATTERY_PROTECT
    g_low_battery_level++;
    if(g_low_battery_level > 2)
    g_low_battery_level = 2; 

    if(g_low_battery_level==1)      
    exec_low_battery_callback(LOW_BATTERY_LEVEL_1);
    else if(g_low_battery_level==2) 
    exec_low_battery_callback(LOW_BATTERY_LEVEL_2);
    else                            

#if 0
    lbat_min_en_setting(0);
    mdelay(1);
    lbat_max_en_setting(1);
#else

    pmic_set_register_value(PMIC_AUXADC_LBAT_VOLT_MIN,BAT_LV_2_THD); 

    lbat_min_en_setting(0);
    lbat_max_en_setting(0);
    mdelay(1);
    if(g_low_battery_level<2)
    {
        lbat_min_en_setting(1);
    }
    lbat_max_en_setting(1);
#endif
#endif 

}

//==============================================================================
// Battery OC call back function
//==============================================================================
#define OCCB_NUM 16

#define DISABLE_BATTERY_OC_PROTECT
#ifndef DISABLE_BATTERY_OC_PROTECT
#define BATTERY_OC_PROTECT
#endif


#ifdef BATTERY_OC_PROTECT
// ex. Ireg = 65535 - (I * 950000uA / 2 / 158.122 / CAR_TUNE_VALUE * 100)
// (950000/2/158.122)*100~=300400

#define BAT_OC_H_THD   65535-((300400*POWER_BAT_OC_CURRENT_H/1000)/CAR_TUNE_VALUE)       //ex: 4670mA
#define BAT_OC_L_THD   65535-((300400*POWER_BAT_OC_CURRENT_L/1000)/CAR_TUNE_VALUE)       //ex: 5500mA

#define BAT_OC_H_THD_RE   65535-((300400*POWER_BAT_OC_CURRENT_H_RE/1000)/CAR_TUNE_VALUE) //ex: 3400mA
#define BAT_OC_L_THD_RE   65535-((300400*POWER_BAT_OC_CURRENT_L_RE/1000)/CAR_TUNE_VALUE) //ex: 4000mA

int g_battery_oc_level=0;
int g_battery_oc_stop=0;

struct battery_oc_callback_table
{
    void *occb;
};

struct battery_oc_callback_table occb_tb[] ={
    {NULL},{NULL},{NULL},{NULL},{NULL},{NULL},{NULL},{NULL},
    {NULL},{NULL},{NULL},{NULL},{NULL},{NULL},{NULL},{NULL}    
};

void (*battery_oc_callback)(BATTERY_OC_LEVEL);

void register_battery_oc_notify( void (*battery_oc_callback)(BATTERY_OC_LEVEL), BATTERY_OC_PRIO prio_val )
{
    occb_tb[prio_val].occb = battery_oc_callback;
}

void exec_battery_oc_callback(BATTERY_OC_LEVEL battery_oc_level) //0:no limit
{
    int i=0;

    if(g_battery_oc_stop==1)
    {
    }
    else
    {
        for(i=0 ; i<OCCB_NUM ; i++) 
        {
            if(occb_tb[i].occb != NULL)
            {
                battery_oc_callback = occb_tb[i].occb;
                battery_oc_callback(battery_oc_level);
            }        
        }
    }
}

void bat_oc_h_en_setting(int en_val)
{
  pmic_set_register_value(PMIC_RG_INT_EN_FG_CUR_H,en_val); // mt6325_upmu_set_rg_int_en_fg_cur_h(en_val);   
}

void bat_oc_l_en_setting(int en_val)
{
   pmic_set_register_value(PMIC_RG_INT_EN_FG_CUR_L,en_val); //mt6325_upmu_set_rg_int_en_fg_cur_l(en_val);
}

void battery_oc_protect_init(void)
{
    pmic_set_register_value(PMIC_FG_CUR_HTH,BAT_OC_H_THD);//mt6325_upmu_set_fg_cur_hth(BAT_OC_H_THD);
    pmic_set_register_value(PMIC_FG_CUR_LTH,BAT_OC_L_THD);//mt6325_upmu_set_fg_cur_lth(BAT_OC_L_THD);

    bat_oc_h_en_setting(0);
    bat_oc_l_en_setting(1);
}

#endif //#ifdef BATTERY_OC_PROTECT

void battery_oc_protect_reinit(void)
{
#ifdef BATTERY_OC_PROTECT
    pmic_set_register_value(PMIC_FG_CUR_HTH,BAT_OC_H_THD_RE);//mt6325_upmu_set_fg_cur_hth(BAT_OC_H_THD_RE);
    pmic_set_register_value(PMIC_FG_CUR_LTH,BAT_OC_L_THD_RE);//mt6325_upmu_set_fg_cur_lth(BAT_OC_L_THD_RE);

#endif
}

void fg_cur_h_int_handler(void)
{
    //sub-task
#ifdef BATTERY_OC_PROTECT
    g_battery_oc_level=0;
    exec_battery_oc_callback(BATTERY_OC_LEVEL_0);
    bat_oc_h_en_setting(0);
    bat_oc_l_en_setting(0);
    mdelay(1);
    bat_oc_l_en_setting(1);
    
#endif 
}

void fg_cur_l_int_handler(void)
{
    //sub-task
#ifdef BATTERY_OC_PROTECT
    g_battery_oc_level=1;
    exec_battery_oc_callback(BATTERY_OC_LEVEL_1);       
    bat_oc_h_en_setting(0);
    bat_oc_l_en_setting(0);
    mdelay(1);
    bat_oc_h_en_setting(1);
#endif
}

//==============================================================================
// 15% notify service
//==============================================================================
static struct hrtimer bat_percent_notify_timer;
static struct task_struct *bat_percent_notify_thread = NULL;
static kal_bool bat_percent_notify_flag = KAL_FALSE;
static DECLARE_WAIT_QUEUE_HEAD(bat_percent_notify_waiter);
struct wake_lock bat_percent_notify_lock;
static DEFINE_MUTEX(bat_percent_notify_mutex);

extern kal_uint32 bat_get_ui_percentage(void);

#define BPCB_NUM 16

#ifndef DISABLE_BATTERY_PERCENT_PROTECT
#define BATTERY_PERCENT_PROTECT
#endif

int g_battery_percent_level=0;
int g_battery_percent_stop=0;

#define BAT_PERCENT_LINIT 15

struct battery_percent_callback_table
{
    void *bpcb;
};

struct battery_percent_callback_table bpcb_tb[] ={
    {NULL},{NULL},{NULL},{NULL},{NULL},{NULL},{NULL},{NULL},
    {NULL},{NULL},{NULL},{NULL},{NULL},{NULL},{NULL},{NULL}    
};

void (*battery_percent_callback)(BATTERY_PERCENT_LEVEL);

void register_battery_percent_notify( void (*battery_percent_callback)(BATTERY_PERCENT_LEVEL), BATTERY_PERCENT_PRIO prio_val )
{
    bpcb_tb[prio_val].bpcb = battery_percent_callback;
    
    if( (g_battery_percent_stop==0) && (g_battery_percent_level==1) )
    {
        battery_percent_callback(BATTERY_PERCENT_LEVEL_1);        
    }
}

void exec_battery_percent_callback(BATTERY_PERCENT_LEVEL battery_percent_level) //0:no limit
{
    int i=0;

    if(g_battery_percent_stop==1)
    {
    }
    else
    {
        for(i=0 ; i<BPCB_NUM ; i++) 
        {
            if(bpcb_tb[i].bpcb != NULL)
            {
                battery_percent_callback = bpcb_tb[i].bpcb;
                battery_percent_callback(battery_percent_level);
            }        
        }
    }
}
#ifdef BATTERY_PERCENT_PROTECT
int bat_percent_notify_handler(void *unused)
{
    ktime_t ktime;
    int bat_per_val=0;

    do
    {
        ktime = ktime_set(10, 0);

        wait_event_interruptible(bat_percent_notify_waiter, (bat_percent_notify_flag == KAL_TRUE));

        wake_lock(&bat_percent_notify_lock);
        mutex_lock(&bat_percent_notify_mutex);

        bat_per_val=bat_get_ui_percentage();
        
        if( (upmu_get_rgs_chrdet()==0) && (g_battery_percent_level==0) && (bat_per_val<=BAT_PERCENT_LINIT) )
        {
            g_battery_percent_level=1;
            exec_battery_percent_callback(BATTERY_PERCENT_LEVEL_1);
        }
        else if( (g_battery_percent_level==1) && (bat_per_val>BAT_PERCENT_LINIT) )
        {
            g_battery_percent_level=0;
            exec_battery_percent_callback(BATTERY_PERCENT_LEVEL_0);
        }
        else
        {
        }
        bat_percent_notify_flag = KAL_FALSE;
        
        mutex_unlock(&bat_percent_notify_mutex);
        wake_unlock(&bat_percent_notify_lock);
       
        hrtimer_start(&bat_percent_notify_timer, ktime, HRTIMER_MODE_REL);    
        
    } while (!kthread_should_stop());
    
    return 0;
}
#endif

#ifdef BATTERY_PERCENT_PROTECT
enum hrtimer_restart bat_percent_notify_task(struct hrtimer *timer)
{
    bat_percent_notify_flag = KAL_TRUE; 
    wake_up_interruptible(&bat_percent_notify_waiter);
    
    return HRTIMER_NORESTART;
}
void bat_percent_notify_init(void)
{
    ktime_t ktime;

    ktime = ktime_set(20, 0);
    hrtimer_init(&bat_percent_notify_timer, CLOCK_MONOTONIC, HRTIMER_MODE_REL);
    bat_percent_notify_timer.function = bat_percent_notify_task;    
    hrtimer_start(&bat_percent_notify_timer, ktime, HRTIMER_MODE_REL);

    bat_percent_notify_thread = kthread_run(bat_percent_notify_handler, 0, "bat_percent_notify_thread");
    if (IS_ERR(bat_percent_notify_thread))
    {
        PMICLOG("Failed to create bat_percent_notify_thread\n");
    }
}
#endif
//==============================================================================
// DLPT service
//==============================================================================
#ifndef DISABLE_DLPT_FEATURE
#define DLPT_FEATURE_SUPPORT
#endif

#ifdef DLPT_FEATURE_SUPPORT
static struct hrtimer dlpt_notify_timer;
static struct task_struct *dlpt_notify_thread = NULL;
static kal_bool dlpt_notify_flag = KAL_FALSE;
static DECLARE_WAIT_QUEUE_HEAD(dlpt_notify_waiter);
struct wake_lock dlpt_notify_lock;
static DEFINE_MUTEX(dlpt_notify_mutex);

#define DLPT_NUM 16


int g_dlpt_stop=0;
unsigned int g_dlpt_val=0;

int g_dlpt_start=0;

int g_iavg_val=0;
int g_imax_val=0;
int g_imax_val_pre=0;

struct dlpt_callback_table
{
    void *dlpt_cb;
};

struct dlpt_callback_table dlpt_cb_tb[] ={
    {NULL},{NULL},{NULL},{NULL},{NULL},{NULL},{NULL},{NULL},
    {NULL},{NULL},{NULL},{NULL},{NULL},{NULL},{NULL},{NULL}    
};
  
void (*dlpt_callback)(unsigned int);

void register_dlpt_notify( void (*dlpt_callback)(unsigned int), DLPT_PRIO prio_val)
{
    dlpt_cb_tb[prio_val].dlpt_cb = dlpt_callback;
    
    if( (g_dlpt_stop==0) && (g_dlpt_val!=0) )
    {
        dlpt_callback(g_dlpt_val);        
    }
}

void exec_dlpt_callback(unsigned int dlpt_val)
{
    int i=0;

    g_dlpt_val = dlpt_val;

    if(g_dlpt_stop==1)
    {
    }
    else
    {
        for(i=0 ; i<DLPT_NUM ; i++) 
        {
            if(dlpt_cb_tb[i].dlpt_cb != NULL)
            {
                dlpt_callback = dlpt_cb_tb[i].dlpt_cb;                
                dlpt_callback(g_dlpt_val);
            }        
        }
    }
}

extern kal_uint32 bat_get_ui_percentage(void);
extern kal_int32 fgauge_read_v_by_d(int d_val);
extern kal_int32 fgauge_read_r_bat_by_v(kal_int32 voltage);

int get_dlpt_iavg(int is_use_zcv)
{
    int bat_cap_val=0;
    int zcv_val=0;        
    int vsys_min_2_val=POWER_INT2_VOLT;
    int rbat_val=0;
    int rdc_val=0;    
    int iavg_val=0;

    bat_cap_val=bat_get_ui_percentage();

    if(is_use_zcv==1)
    {
        zcv_val=fgauge_read_v_by_d(100-bat_cap_val);
    }
    else
    {
        #if defined(SWCHR_POWER_PATH)
        zcv_val=PMIC_IMM_GetOneChannelValue(MT6328_AUX_ISENSE_AP,5,1);  
        #else
        zcv_val=PMIC_IMM_GetOneChannelValue(MT6328_AUX_BATSNS_AP,5,1);
        #endif
    }
    rbat_val=fgauge_read_r_bat_by_v(zcv_val);
    rdc_val=CUST_R_FG_OFFSET+R_FG_VALUE+rbat_val;

    if(rdc_val==0) 
        rdc_val=1;    
    iavg_val=((zcv_val-vsys_min_2_val)*1000)/rdc_val; // ((mV)*1000)/m-ohm
    
    return iavg_val;
}

int get_real_volt(int val) //0.1mV
{
    int ret=0;
    ret = val&0x7FFF; 
    ret = (ret*4*1800*10)/32768; 
    return ret;
}

int get_real_curr(int val) //0.1mA
{
    int ret=0;

    if( val > 32767 ) // > 0x8000
    {
        ret = val-65535;
        ret = ret-(ret*2);
    }
    else
    {
        ret = val;
    }
    ret = ret*158122;    
    do_div(ret, 100000);
    ret = (ret*20)/R_FG_VALUE;
    ret = ((ret*CAR_TUNE_VALUE)/100);

    return ret;
}

int get_rac_val(void)
{
#if 0
    PMICLOG("[get_rac_val] TBD\n");
    return 1;
#else
    int volt_1=0;
    int volt_2=0;
    int curr_1=0;
    int curr_2=0;
    int rac_cal=0;
    int ret=0;
    kal_bool retry_state = KAL_FALSE;
    int retry_count=0;

    do
    {        
        //adc and fg--------------------------------------------------------
        #if defined(SWCHR_POWER_PATH)
        volt_1=PMIC_IMM_GetOneChannelValue(MT6328_AUX_ISENSE_AP,5,10);  
        #else
        volt_1=PMIC_IMM_GetOneChannelValue(MT6328_AUX_BATSNS_AP,5,10);
        #endif
        curr_1=battery_meter_get_battery_current();    
        
        //Enable dummy load--------------------------------------------------
        pmic_config_interface(0x23C,0x40,0xFFFF,0); //Reg[0x238][6]=0
        //mt6325_upmu_set_rg_g_drv_2m_ck_pdn(0);
        //mt6325_upmu_set_rg_drv_32k_ck_pdn(0);
        //mt6325_upmu_set_rg_drv_isink2_ck_cksel(0);
        //mt6325_upmu_set_isink_ch2_mode(3);
        //mt6325_upmu_set_isink_chop2_en(1);
        //mt6325_upmu_set_isink_ch2_bias_en(1);
        //mt6325_upmu_set_isink_ch2_step(7);
        //mt6325_upmu_set_rg_isink2_double_en(1);
        //mt6325_upmu_set_isink_ch2_en(1);


        //Wait --------------------------------------------------------------
        msleep(50);
        
        //adc and fg--------------------------------------------------------
        #if defined(SWCHR_POWER_PATH)
        volt_2=PMIC_IMM_GetOneChannelValue(MT6328_AUX_ISENSE_AP,5,10);  
        #else
        volt_2=PMIC_IMM_GetOneChannelValue(MT6328_AUX_BATSNS_AP,5,10);
        #endif
        curr_2=battery_meter_get_battery_current();    
        
        //Disable dummy load-------------------------------------------------
        //mt6325_upmu_set_isink_ch2_en(0);
        ////pmic_config_interface(0x23A,0x40,0xFFFF,0); //Reg[0x238][6]=1, move to suspend callback
                
        //Calculate Rac------------------------------------------------------
        if( (curr_2-curr_1) >= 400 ) //40.0mA
        {
            rac_cal=((volt_1-volt_2)*1000)/(curr_2-curr_1); //m-ohm

            if(rac_cal<0) ret = rac_cal-(rac_cal*2);
            else          ret = rac_cal;
        }
        else if( (curr_1-curr_2) >= 400 ) //40.0mA
        {
            rac_cal=((volt_2-volt_1)*1000)/(curr_1-curr_2); //m-ohm

            if(rac_cal<0) ret = rac_cal-(rac_cal*2);
            else          ret = rac_cal;
        }
        else
        {
            ret=-1;
        }  

        //------------------------
        retry_count++;
        
        if((retry_count < 3) && (ret == -1)) retry_state = KAL_TRUE;
        else                                 retry_state = KAL_FALSE;

    }while(retry_state == KAL_TRUE);
      
    return ret;
#endif    
}

int get_dlpt_imax(void)
{
    int bat_cap_val=0;
    int zcv_val=0;        
    int vsys_min_1_val=POWER_INT1_VOLT;
    int rac_val=0;    
    int imax_val=0;

    bat_cap_val=bat_get_ui_percentage();
    zcv_val=fgauge_read_v_by_d(100-bat_cap_val);

    rac_val=get_rac_val();
    
    if(rac_val==-1)
        return -1;

    if(rac_val==0) 
        rac_val=1;
    imax_val=((zcv_val-vsys_min_1_val)*1000)/rac_val; // ((mV)*1000)/m-ohm
    
    return imax_val;
}

int dlpt_check_power_off(void)
{   
    int ret=0;

    ret=0;
    if(g_dlpt_start==0)
    {
    }
    else
    {
        if (!upmu_get_rgs_chrdet())
        {
            if( (g_iavg_val<POWEROFF_BAT_CURRENT) && (g_low_battery_level>0) ) {
                ret=1;        
            } else if(g_low_battery_level==2) {
                ret=1;
            } else {
                ret=0;
            }
        }
    }

    return ret;    
}

int get_dlpt_imax_avg(void)
{
    int i=0;
    int val[5]={0,0,0,0,0};
    int val_compare=g_iavg_val*3;
    int timeout_i=0;    
    int timeout_val=100;    

    for(i=0;i<5;i++)
    {
        val[i]=get_dlpt_imax();
        if( (val[i]>val_compare) || (val[i]<g_iavg_val) || (val[i]<=0) ) 
        { 
            i--; //re-measure
        }
        
        timeout_i++;
        if(timeout_i>=timeout_val)
        {
            return -1;
        }
    }
    
    if(i<=0) 
        return -1;
    else     
        return (val[0]+val[1]+val[2]+val[3]+val[4])/5;
}

int dlpt_notify_handler(void *unused)
{
    ktime_t ktime;
    int pre_ui_soc=0;
    int cur_ui_soc=0;
    int diff_ui_soc=1;
    pre_ui_soc=bat_get_ui_percentage();
    cur_ui_soc=pre_ui_soc;
    
    do
    {
        ktime = ktime_set(10, 0);

        wait_event_interruptible(dlpt_notify_waiter, (dlpt_notify_flag == KAL_TRUE));

        wake_lock(&dlpt_notify_lock);
        mutex_lock(&dlpt_notify_mutex);
        //---------------------------------

        cur_ui_soc=bat_get_ui_percentage();
        //if(1)
        if( ((pre_ui_soc-cur_ui_soc)>=diff_ui_soc) || 
            ((cur_ui_soc-pre_ui_soc)>=diff_ui_soc) ||
            (g_imax_val == 0)                      ||
            (g_imax_val == -1)
          )
        
        {   
            if (upmu_get_rgs_chrdet())
            {   
                g_iavg_val=get_dlpt_iavg(0);
                g_imax_val=(g_iavg_val*125)/100;
                
                //sync value
                g_imax_val_pre=g_imax_val;                
            }
            else
            {
                g_iavg_val=get_dlpt_iavg(1);
                g_imax_val=get_dlpt_imax_avg();
                
                //sync value
                if(g_imax_val != -1)
                {
                    if(g_imax_val_pre <= 0)
                    {
                        g_imax_val_pre=g_imax_val;
                    }
                    if(g_imax_val > g_imax_val_pre)
                    {
                        g_imax_val=g_imax_val_pre;
                    }
                    else
                    {
                        g_imax_val_pre=g_imax_val;
                    }
                }
            }
            
            //10/21, Ricky, report imax=iavg*1.25
            //g_imax_val=(g_iavg_val*125)/100;
            //PMICLOG("[DLPT] Ricky(imax=%d,iavg=%d)\n", g_imax_val, g_iavg_val);
            
            //Notify
            if(g_imax_val != -1)
            {
                if(g_imax_val > IMAX_MAX_VALUE)
                {                    
                    g_imax_val = IMAX_MAX_VALUE;
                }
                exec_dlpt_callback(g_imax_val);                
            }

            pre_ui_soc=cur_ui_soc;
        }
        
        g_dlpt_start=1;
        dlpt_notify_flag = KAL_FALSE;

        //---------------------------------
        mutex_unlock(&dlpt_notify_mutex);
        wake_unlock(&dlpt_notify_lock);
       
        hrtimer_start(&dlpt_notify_timer, ktime, HRTIMER_MODE_REL);    
        
    } while (!kthread_should_stop());
    
    return 0;
}

enum hrtimer_restart dlpt_notify_task(struct hrtimer *timer)
{
    dlpt_notify_flag = KAL_TRUE; 
    wake_up_interruptible(&dlpt_notify_waiter);
    
    return HRTIMER_NORESTART;
}

int get_system_loading_ma(void)
{
    int fg_val=0;
    
    if(g_dlpt_start==0) {    
    } else {
        fg_val = battery_meter_get_battery_current();
        fg_val = fg_val/10;
        if(battery_meter_get_battery_current_sign()==1)
        {
            fg_val = 0-fg_val; // charging
        }        
    }
    
    return fg_val;
}


void dlpt_notify_init(void)
{
    ktime_t ktime;

    ktime = ktime_set(30, 0);
    hrtimer_init(&dlpt_notify_timer, CLOCK_MONOTONIC, HRTIMER_MODE_REL);
    dlpt_notify_timer.function = dlpt_notify_task;    
    hrtimer_start(&dlpt_notify_timer, ktime, HRTIMER_MODE_REL);

    dlpt_notify_thread = kthread_run(dlpt_notify_handler, 0, "dlpt_notify_thread");

pmic_set_register_value(PMIC_RG_UVLO_VTHL,0);

    //re-init UVLO volt
    switch(POWER_UVLO_VOLT_LEVEL){       
        case 2500: pmic_set_register_value(PMIC_RG_UVLO_VTHL,0); break;
        case 2550: pmic_set_register_value(PMIC_RG_UVLO_VTHL,1); break;
        case 2600: pmic_set_register_value(PMIC_RG_UVLO_VTHL,2); break;
        case 2650: pmic_set_register_value(PMIC_RG_UVLO_VTHL,3); break;
        case 2700: pmic_set_register_value(PMIC_RG_UVLO_VTHL,4);break;
        case 2750: pmic_set_register_value(PMIC_RG_UVLO_VTHL,5); break;
        case 2800: pmic_set_register_value(PMIC_RG_UVLO_VTHL,6); break;
        case 2850: pmic_set_register_value(PMIC_RG_UVLO_VTHL,7); break;
        case 2900: pmic_set_register_value(PMIC_RG_UVLO_VTHL,8); break;        
        default:
            PMICLOG("Invalid value(%d)\n", POWER_UVLO_VOLT_LEVEL);
            break;
    }    
}

#endif //#ifdef DLPT_FEATURE_SUPPORT

//==============================================================================
// interrupt Setting 
//==============================================================================
static struct pmic_interrupt_bit interrupt_status0[] = {
    PMIC_S_INT_GEN(RG_INT_STATUS_SPKL_AB),
    PMIC_S_INT_GEN(RG_INT_STATUS_SPKL),
    PMIC_S_INT_GEN(RG_INT_STATUS_BAT_L),
    PMIC_S_INT_GEN(RG_INT_STATUS_BAT_H),
    PMIC_S_INT_GEN(RG_INT_STATUS_WATCHDOG),
    PMIC_S_INT_GEN(RG_INT_STATUS_PWRKEY),
    PMIC_S_INT_GEN(RG_INT_STATUS_THR_L),
    PMIC_S_INT_GEN(RG_INT_STATUS_THR_H),
    PMIC_S_INT_GEN(RG_INT_STATUS_VBATON_UNDET),
    PMIC_S_INT_GEN(RG_INT_STATUS_BVALID_DET),
    PMIC_S_INT_GEN(RG_INT_STATUS_CHRDET),
    PMIC_S_INT_GEN(RG_INT_STATUS_OV),
    PMIC_S_INT_GEN(NO_USE),
    PMIC_S_INT_GEN(NO_USE),
    PMIC_S_INT_GEN(NO_USE),
    PMIC_S_INT_GEN(NO_USE),
};

static struct pmic_interrupt_bit interrupt_status1[] = {
    PMIC_S_INT_GEN(RG_INT_STATUS_LDO),
    PMIC_S_INT_GEN(RG_INT_STATUS_FCHRKEY),
    PMIC_S_INT_GEN(RG_INT_STATUS_ACCDET),
    PMIC_S_INT_GEN(RG_INT_STATUS_AUDIO),
    PMIC_S_INT_GEN(RG_INT_STATUS_RTC),
    PMIC_S_INT_GEN(RG_INT_STATUS_VPROC),
    PMIC_S_INT_GEN(RG_INT_STATUS_VSYS),
    PMIC_S_INT_GEN(RG_INT_STATUS_VPA),
    PMIC_S_INT_GEN(NO_USE),
    PMIC_S_INT_GEN(NO_USE),
    PMIC_S_INT_GEN(NO_USE),
    PMIC_S_INT_GEN(NO_USE),
    PMIC_S_INT_GEN(NO_USE),
    PMIC_S_INT_GEN(NO_USE),
    PMIC_S_INT_GEN(NO_USE),
    PMIC_S_INT_GEN(NO_USE),
};



static struct pmic_interrupts interrupts[] = {
    PMIC_M_INTS_GEN(MT6350_INT_STATUS0,MT6350_INT_CON0,MT6350_INT_CON0_SET,MT6350_INT_CON0_CLR,interrupt_status0),
    PMIC_M_INTS_GEN(MT6350_INT_STATUS1,MT6350_INT_CON1,MT6350_INT_CON1_SET,MT6350_INT_CON1_CLR,interrupt_status1),
};

extern void kpd_pwrkey_pmic_handler(unsigned long pressed);
void pwrkey_int_handler(void)
{
    kal_uint32 ret=0;
    static kal_bool  key_press = KAL_FALSE;
    
    if(pmic_get_register_value(PMIC_PWRKEY_DEB) == 1)
    {
#if defined (CONFIG_MTK_KERNEL_POWER_OFF_CHARGING)
        if(g_boot_mode == KERNEL_POWER_OFF_CHARGING_BOOT)
        {
            if(key_press == KAL_TRUE)
            {
                key_press = KAL_FALSE;
                timer_pos = sched_clock();
                if(timer_pos - timer_pre >= LONG_PWRKEY_PRESS_TIME)
                {
                    long_pwrkey_press = true;
                }
                if(long_pwrkey_press)   //500ms
                {
                    arch_reset(0, NULL);
                }
            }
        }
#endif
         kpd_pwrkey_pmic_handler(0x0);
        pmic_set_register_value(PMIC_RG_PWRKEY_INT_SEL, 0);
    }
    else
    {
#if defined (CONFIG_MTK_KERNEL_POWER_OFF_CHARGING)
        if(g_boot_mode == KERNEL_POWER_OFF_CHARGING_BOOT)
        {
            key_press = KAL_TRUE;
            timer_pre = sched_clock();
        }
#endif
        kpd_pwrkey_pmic_handler(0x1);
        pmic_set_register_value(PMIC_RG_PWRKEY_INT_SEL, 1);;
    }
   // ret = pmic_config_interface(MT6350_INT_STATUS0, 0x1, 0x1, 0x5);  
}


//==============================================================================
// PMIC Interrupt callback
//==============================================================================
#ifdef CONFIG_MTK_KEYPAD
extern void kpd_pmic_rstkey_handler(unsigned long pressed);
#else
void kpd_pmic_rstkey_handler(unsigned long pressed)
{
}
#endif
void homekey_int_handler(void)
{
    kal_uint32 ret=0;

#ifdef KPD_PMIC_RSTKEY_MAP
    if(pmic_get_register_value(PMIC_FCHRKEY_DEB)==1)
    {
        kpd_pmic_rstkey_handler(0x0);
    }
    else
    {
        kpd_pmic_rstkey_handler(0x1);
    }    
#else    
#endif
    
    //ret=pmic_config_interface(MT6350_INT_STATUS1, 0x1, 0x1, 0x1);        
    //kpd_pmic_rstkey_handler(0x1);   
}

void chrdet_int_handler(void)
{
#ifdef CONFIG_MTK_KERNEL_POWER_OFF_CHARGING
    if (!upmu_get_rgs_chrdet())
    {
        int boot_mode = 0;
        boot_mode = get_boot_mode();
        
        if(boot_mode == KERNEL_POWER_OFF_CHARGING_BOOT || boot_mode == LOW_POWER_OFF_CHARGING_BOOT)
        {
            mt_power_off();
        }
    }
#endif
    pmic_set_register_value(PMIC_RG_USBDL_RST,1);
#ifdef CONFIG_POWER_EXT
#else
    do_chrdet_int_task();
#endif
}

#ifdef CONFIG_MTK_ACCDET
extern int accdet_irq_handler(void);
void accdet_int_handler(void)
{
    kal_uint32 ret=0;

    
    ret = accdet_irq_handler();
	if(0 == ret){
	}
//    ret=pmic_config_interface(INT_STATUS1,0x1,0x1,2);    
}
#endif

void rtc_int_handler(void)
{
    kal_uint32 ret=0;

#ifndef CONFIG_EARLY_LINUX_PORTING
    rtc_irq_handler();
#endif
    //ret=pmic_config_interface(MT6350_INT_STATUS1, 0x1, 0x1, 4);    
}

void auxadc_imp_int_handler_r(void)
{/*
    PMICLOG("auxadc_imp_int_handler_r() =%d \n",pmic_get_register_value(PMIC_AUXADC_ADC_OUT_IMP));
    //clear IRQ
    pmic_set_register_value(PMIC_AUXADC_CLR_IMP_CNT_STOP,1);
    pmic_set_register_value(PMIC_AUXADC_IMPEDANCE_IRQ_CLR,1);   

    //restore to initial state
    pmic_set_register_value(PMIC_AUXADC_CLR_IMP_CNT_STOP,0);
    pmic_set_register_value(PMIC_AUXADC_IMPEDANCE_IRQ_CLR,0);       

    //turn off interrupt
    pmic_set_register_value(PMIC_RG_INT_EN_AUXADC_IMP,0);
*/
}

//==============================================================================
// PMIC Interrupt service
//==============================================================================
static DEFINE_MUTEX(pmic_mutex);
static struct task_struct *pmic_thread_handle = NULL;
struct wake_lock pmicThread_lock;

void wake_up_pmic(void)
{
    wake_up_process(pmic_thread_handle);
    wake_lock(&pmicThread_lock);    
}
EXPORT_SYMBOL(wake_up_pmic);

void mt_pmic_eint_irq(void)
{
    wake_up_pmic();
    return ;
}

void pmic_enable_interrupt(kal_uint32 intNo,kal_uint32 en,char *str)
{
    kal_uint32 shift,no;
    shift=intNo/PMIC_INT_WIDTH;
    no=intNo%PMIC_INT_WIDTH;

    if(shift >=ARRAY_SIZE(interrupts))
    {
        return;
    }

    if(en==1)
    {
        pmic_config_interface(interrupts[shift].set,0x1,0x1,no);  
    }
    else if(en==0)
    {
        pmic_config_interface(interrupts[shift].clear,0x1,0x1,no);  
    }
} 

void pmic_register_interrupt_callback(kal_uint32 intNo,void (EINT_FUNC_PTR)(void))
{
    kal_uint32 shift,no;
    shift=intNo/PMIC_INT_WIDTH;
    no=intNo%PMIC_INT_WIDTH;

    if(shift >=ARRAY_SIZE(interrupts))
    {
        PMICLOG("[pmic_register_interrupt_callback] fail intno=%d \r\n",intNo);
        return;
    }

    interrupts[shift].interrupts[no].callback=EINT_FUNC_PTR;

}

void PMIC_EINT_SETTING(void)
{
        upmu_set_reg_value(MT6350_INT_CON0, 0); 
        upmu_set_reg_value(MT6350_INT_CON1, 0);
        
        // Enable pwrkey/homekey interrupt
        upmu_set_reg_value(MT6350_INT_CON0_SET, 0x0420); // bit[10], bit[5]
        upmu_set_reg_value(MT6350_INT_CON1_SET, 0x0016); // bit[4], bit[2], bit[1]
        
        //for all interrupt events, turn on interrupt module clock
        pmic_set_register_value(PMIC_RG_INTRP_CK_PDN,0);    
        
        pmic_register_interrupt_callback(5,pwrkey_int_handler);
        pmic_register_interrupt_callback(10,chrdet_int_handler);
        pmic_register_interrupt_callback(17,homekey_int_handler);
#ifdef CONFIG_MTK_ACCDET
        pmic_register_interrupt_callback(18,accdet_int_handler);
#endif
        pmic_register_interrupt_callback(20,rtc_int_handler);
        
        pmic_enable_interrupt(5,1,"PMIC");
        pmic_enable_interrupt(10,1,"PMIC");
        pmic_enable_interrupt(17,1,"PMIC");
#ifdef CONFIG_MTK_ACCDET
        pmic_enable_interrupt(18,1,"PMIC");
#endif
        pmic_enable_interrupt(20,1,"PMIC");
        mt_eint_set_hw_debounce(g_eint_pmic_num, g_cust_eint_mt_pmic_debounce_cn);    
        mt_eint_registration(g_eint_pmic_num,g_cust_eint_mt_pmic_type,mt_pmic_eint_irq,0);
        mt_eint_unmask(g_eint_pmic_num);    
}

static void pmic_int_handler(void)
{
        kal_uint8 i,j;
        kal_uint32 ret;
        kal_uint32 int0, int1;    

#if 1
        int0 = upmu_get_reg_value(MT6350_INT_CON0);
        int1 = upmu_get_reg_value(MT6350_INT_CON1);
        upmu_set_reg_value(MT6350_INT_CON0_CLR, 0x0210); // bit[9], bit[4]    
    
#endif    
        for (i = 0; i < ARRAY_SIZE(interrupts); i++) 
        {
                kal_uint32 int_status_val=0;
                int_status_val=upmu_get_reg_value(interrupts[i].address);
        
                for(j=0;j<PMIC_INT_WIDTH;j++)
                {
                        if((int_status_val)&(1<<j))
                        {
                                if (interrupts[i].interrupts[j].callback!=NULL)
                                {
                                        interrupts[i].interrupts[j].callback();
                                        interrupts[i].interrupts[j].times++;
                                }
                                ret=pmic_config_interface(interrupts[i].address,0x1,0x1,j);    
                        }
                }
        }   
}

static int pmic_thread_kthread(void *x)
{
    kal_uint32 i;
    kal_uint32 int_status_val=0;
    U32 pwrap_eint_status=0;
    struct sched_param param = { .sched_priority = 98 };

    sched_setscheduler(current, SCHED_FIFO, &param);
    set_current_state(TASK_INTERRUPTIBLE);

    /* Run on a process content */
    while (1) {
        mutex_lock(&pmic_mutex);

        //pwrap_eint_status = pmic_wrap_eint_status();    
        //PMICLOG("[PMIC_INT] pwrap_eint_status=0x%x\n", pwrap_eint_status);

        pmic_int_handler();

        //pmic_wrap_eint_clr(0x0);
        //PMICLOG("[PMIC_INT] pmic_wrap_eint_clr(0x0);\n");

        for (i = 0; i < ARRAY_SIZE(interrupts); i++) 
        {
            int_status_val=upmu_get_reg_value(interrupts[i].address);

        }   

        mdelay(1);

        mutex_unlock(&pmic_mutex);
        wake_unlock(&pmicThread_lock);

        set_current_state(TASK_INTERRUPTIBLE);        
        mt_eint_unmask(g_eint_pmic_num);  //need fix
        schedule();
    }

    return 0;
}

int pmic_is_ext_buck_exist(void)
{
        return 0;    
}

int is_ext_buck2_exist(void)
{
//        return mt_get_gpio_in(140);
    return 0;
}

int is_ext_vbat_boost_exist(void)
{
        return 0;    
}

int is_ext_swchr_exist(void)
{
    return 0;
}

/*
//==============================================================================
// Enternal SWCHR
//==============================================================================
#ifdef MTK_BQ24261_SUPPORT
extern int is_bq24261_exist(void);
#endif

int is_ext_swchr_exist(void)
{
    #ifdef MTK_BQ24261_SUPPORT
        if( (is_bq24261_exist()==1) )
            return 1;    
        else
            return 0;    
    #else
        PMICLOG("[is_ext_swchr_exist] no define any HW\n");
        return 0;
    #endif    
}


//==============================================================================
// Enternal VBAT Boost status
//==============================================================================
extern int is_tps6128x_sw_ready(void);
extern int is_tps6128x_exist(void);

int is_ext_vbat_boost_sw_ready(void)
{    
    if( (is_tps6128x_sw_ready()==1) )
        return 1;    
    else
        return 0;    
}

int is_ext_vbat_boost_exist(void)
{
    if( (is_tps6128x_exist()==1) )
        return 1;
    else
        return 0;    
}


//==============================================================================
// Enternal BUCK status
//==============================================================================
extern int is_mt6311_sw_ready(void);
extern int is_mt6311_exist(void);
extern int get_mt6311_i2c_ch_num(void);

int get_ext_buck_i2c_ch_num(void)
{
    if(is_mt6311_exist()==1)
    {
        return get_mt6311_i2c_ch_num();
    }
    else
    {
        return -1;
    }
}

int is_ext_buck_sw_ready(void)
{    
    if( (is_mt6311_sw_ready()==1) )
        return 1;    
    else
        return 0;    
}

int pmic_is_ext_buck_exist(void)
{
    if( (is_mt6311_exist()==1) )
        return 1;
    else
        return 0;    
}
*/
//==============================================================================
// FTM 
//==============================================================================
#define PMIC_DEVNAME "pmic_ftm"
#define Get_IS_EXT_BUCK_EXIST _IOW('k', 20, int)
#define Get_IS_EXT_VBAT_BOOST_EXIST _IOW('k', 21, int)
#define Get_IS_EXT_SWCHR_EXIST _IOW('k', 22, int)
#define Get_IS_EXT_BUCK2_EXIST _IOW('k', 23, int)


static struct class *pmic_class = NULL;
static struct cdev *pmic_cdev;
static int pmic_major = 0;
static dev_t pmic_devno;

static long pmic_ftm_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
    int *user_data_addr;
    int ret = 0;
    int adc_in_data[2] = {1,1};
    int adc_out_data[2] = {1,1};

    switch(cmd)
    {
        //#if defined(FTM_EXT_BUCK_CHECK)
            case Get_IS_EXT_BUCK_EXIST:
                user_data_addr = (int *)arg;
                ret = copy_from_user(adc_in_data, user_data_addr, 8);
                adc_out_data[0] = pmic_is_ext_buck_exist();
                ret = copy_to_user(user_data_addr, adc_out_data, 8); 
            break;
        //#endif

        //#if defined(FTM_EXT_VBAT_BOOST_CHECK)
            case Get_IS_EXT_VBAT_BOOST_EXIST:
                user_data_addr = (int *)arg;
                ret = copy_from_user(adc_in_data, user_data_addr, 8);
                adc_out_data[0] = is_ext_vbat_boost_exist();
                ret = copy_to_user(user_data_addr, adc_out_data, 8); 
            break;
        //#endif
        
        //#if defined(FEATURE_FTM_SWCHR_HW_DETECT)
            case Get_IS_EXT_SWCHR_EXIST:
                user_data_addr = (int *)arg;
                ret = copy_from_user(adc_in_data, user_data_addr, 8);
                adc_out_data[0] = is_ext_swchr_exist();
                ret = copy_to_user(user_data_addr, adc_out_data, 8); 
            break;
        //#endif
            case Get_IS_EXT_BUCK2_EXIST:
                user_data_addr = (int *)arg;
                ret = copy_from_user(adc_in_data, user_data_addr, 8);
                adc_out_data[0] = is_ext_buck2_exist();
                ret = copy_to_user(user_data_addr, adc_out_data, 8); 
            break;        
        default:
            break;
    }
    
    return 0;
}

static int pmic_ftm_open(struct inode *inode, struct file *file)
{ 
   return 0;
}

static int pmic_ftm_release(struct inode *inode, struct file *file)
{
    return 0;
}


static struct file_operations pmic_ftm_fops = {
    .owner          = THIS_MODULE,
    .unlocked_ioctl = pmic_ftm_ioctl,
    .open           = pmic_ftm_open,
    .release        = pmic_ftm_release,    
};

void pmic_ftm_init(void)
{
    struct class_device *class_dev = NULL;
    int ret=0;
    
    ret = alloc_chrdev_region(&pmic_devno, 0, 1, PMIC_DEVNAME);
    if (ret) 
        PMICLOG("[pmic_ftm_init] Error: Can't Get Major number for pmic_ftm\n");
    
    pmic_cdev = cdev_alloc();
    pmic_cdev->owner = THIS_MODULE;
    pmic_cdev->ops = &pmic_ftm_fops;

    ret = cdev_add(pmic_cdev, pmic_devno, 1);
    if(ret)
        PMICLOG("[pmic_ftm_init] Error: cdev_add\n");
    
    pmic_major = MAJOR(pmic_devno);
    pmic_class = class_create(THIS_MODULE, PMIC_DEVNAME);
    
    class_dev = (struct class_device *)device_create(pmic_class, 
                                                   NULL, 
                                                   pmic_devno, 
                                                   NULL, 
                                                   PMIC_DEVNAME);
}


//==============================================================================
// HW Setting 
//==============================================================================
void PMIC_INIT_SETTING_V1(void)
{
        U32 chip_version = 0;
        U32 ret = 0;
        
        chip_version = pmic_get_register_value(PMIC_CID);
        
        //--------------------------------------------------------
        
        ret = pmic_config_interface(0x2,0xB,0xF,4); // [7:4]: RG_VCDT_HV_VTH; 7V OVP, Tim,Zax
        ret = pmic_config_interface(0xC,0x1,0x7,1); // [3:1]: RG_VBAT_OV_VTH; VBAT_OV=4.3V, Tim Zax, need to check Description
        ret = pmic_config_interface(0x1A,0x3,0xF,0); // [3:0]: RG_CHRWDT_TD; align to 6250's, Tim,zax
        ret = pmic_config_interface(0x20,0x1,0x1,2); // [2:2]: RG_USBDL_RST; Maxwell
        ret = pmic_config_interface(0x24,0x1,0x1,1); // [1:1]: RG_BC11_RST; Reset BC11 detection, Tim,zax,charger programming guide
        ret = pmic_config_interface(0x2A,0x0,0x7,4); // [6:4]: RG_CSDAC_STP; align to 6250's setting,Tim, Zax
        ret = pmic_config_interface(0x2E,0x1,0x1,2); // [2:2]: RG_CSDAC_MODE; Tim,Zax
        ret = pmic_config_interface(0x2E,0x0,0x1,6); // [6:6]: RG_HWCV_EN; Align to MT6328,Top off set=1, TIM,Zax
        ret = pmic_config_interface(0x2E,0x1,0x1,7); // [7:7]: RG_ULC_DET_EN; Tim,Zax
        ret = pmic_config_interface(0x40,0x1,0x1,4); // [4:4]: RG_EN_DRVSEL; Ricky
        ret = pmic_config_interface(0x40,0x1,0x1,5); // [5:5]: RG_RST_DRVSEL; Ricky
        ret = pmic_config_interface(0x46,0x1,0x1,1); // [1:1]: PWRBB_DEB_EN; Ricky
        ret = pmic_config_interface(0x48,0x1,0x1,8); // [8:8]: VPROC_PG_H2L_EN; Ricky
        ret = pmic_config_interface(0x48,0x1,0x1,9); // [9:9]: VSYS_PG_H2L_EN; Ricky
        ret = pmic_config_interface(0x4E,0x1,0x1,5); // [5:5]: STRUP_AUXADC_RSTB_SW; need to check with ABB
        ret = pmic_config_interface(0x4E,0x1,0x1,7); // [7:7]: STRUP_AUXADC_RSTB_SEL; need to check with ABB
        ret = pmic_config_interface(0x50,0x1,0x1,0); // [0:0]: STRUP_PWROFF_SEQ_EN; Ricky
        ret = pmic_config_interface(0x50,0x1,0x1,1); // [1:1]: STRUP_PWROFF_PREOFF_EN; Ricky
        ret = pmic_config_interface(0x52,0x1,0x1,9); // [9:9]: SPK_THER_SHDN_L_EN; 
        ret = pmic_config_interface(0x56,0x1,0x1,0); // [0:0]: RG_SPK_INTG_RST_L; 
        ret = pmic_config_interface(0x64,0x1,0xF,8); // [11:8]: RG_SPKPGA_GAIN; 
        ret = pmic_config_interface(0x102,0x1,0x1,6); // [6:6]: RG_RTC_75K_CK_PDN; JT
        ret = pmic_config_interface(0x102,0x0,0x1,11); // [11:11]: RG_DRV_32K_CK_PDN; JT
        ret = pmic_config_interface(0x102,0x1,0x1,15); // [15:15]: RG_BUCK32K_PDN; JT
        ret = pmic_config_interface(0x108,0x1,0x1,12); // [12:12]: RG_EFUSE_CK_PDN; JT
        ret = pmic_config_interface(0x10E,0x1,0x1,5); // [5:5]: RG_AUXADC_CTL_CK_PDN; JT
        ret = pmic_config_interface(0x120,0x1,0x1,4); // [4:4]: RG_SRCLKEN_HW_MODE; JT
        ret = pmic_config_interface(0x120,0x1,0x1,5); // [5:5]: RG_OSC_HW_MODE; JT
        ret = pmic_config_interface(0x130,0x1,0xFF,8); // [15:8]: RG_TOP_CKTST2_RSV_15_8; JT
        ret = pmic_config_interface(0x148,0x1,0x1,0); // [0:0]: RG_SMT_SYSRSTB; Ricky
        ret = pmic_config_interface(0x148,0x1,0x1,1); // [1:1]: RG_SMT_INT; Ricky
        ret = pmic_config_interface(0x148,0x1,0x1,2); // [2:2]: RG_SMT_SRCLKEN; Ricky
        ret = pmic_config_interface(0x148,0x1,0x1,3); // [3:3]: RG_SMT_RTC_32K1V8; Ricky
        ret = pmic_config_interface(0x160,0x1,0x1,2); // [2:2]: RG_INT_EN_BAT_L; Ricky
        ret = pmic_config_interface(0x160,0x1,0x1,6); // [6:6]: RG_INT_EN_THR_L; Ricky
        ret = pmic_config_interface(0x160,0x1,0x1,7); // [7:7]: RG_INT_EN_THR_H; Ricky
        ret = pmic_config_interface(0x166,0x1,0x1,1); // [1:1]: RG_INT_EN_FCHRKEY; Ricky
        ret = pmic_config_interface(0x212,0x2,0x3,4); // [5:4]: QI_VPROC_VSLEEP; TH
        ret = pmic_config_interface(0x216,0x1,0x1,1); // [1:1]: VPROC_VOSEL_CTRL; Seven,JT
        ret = pmic_config_interface(0x21C,0x17,0x7F,0); // [6:0]: VPROC_SFCHG_FRATE; Seven,JT, the value should setting before EN
        ret = pmic_config_interface(0x21C,0x1,0x1,7); // [7:7]: VPROC_SFCHG_FEN; Seven,JT
        ret = pmic_config_interface(0x21C,0x1,0x1,15); // [15:15]: VPROC_SFCHG_REN; Seven,JT
        ret = pmic_config_interface(0x21E,0x38,0x7F,0); // [6:0]: VPROC_VOSEL; for Low power, the value should after VPROC_VOSEL_CTRL
        ret = pmic_config_interface(0x222,0x18,0x7F,0); // [6:0]: VPROC_VOSEL_SLEEP; Seven, set value same to Vsleep
        ret = pmic_config_interface(0x230,0x3,0x3,0); // [1:0]: VPROC_TRANSTD; Seven,same with 6323
        ret = pmic_config_interface(0x230,0x1,0x3,4); // [5:4]: VPROC_VOSEL_TRANS_EN; Seven,need to recheck
        ret = pmic_config_interface(0x230,0x1,0x1,8); // [8:8]: VPROC_VSLEEP_EN; Seven,JT
        ret = pmic_config_interface(0x238,0x3,0x3,0); // [1:0]: RG_VSYS_SLP; Johnson,same 6323
        ret = pmic_config_interface(0x238,0x2,0x3,4); // [5:4]: QI_VSYS_VSLEEP; Johnson,re-confrim
        ret = pmic_config_interface(0x242,0x23,0x7F,0); // [6:0]: VSYS_SFCHG_FRATE; Johnson,same 6323
        ret = pmic_config_interface(0x242,0x11,0x7F,8); // [14:8]: VSYS_SFCHG_RRATE; Johnson,same 6323
        ret = pmic_config_interface(0x242,0x1,0x1,15); // [15:15]: VSYS_SFCHG_REN; Johnson,same 6323
        ret = pmic_config_interface(0x256,0x3,0x3,0); // [1:0]: VSYS_TRANSTD; Johnson,same 6323
        ret = pmic_config_interface(0x256,0x1,0x3,4); // [5:4]: VSYS_VOSEL_TRANS_EN; Johnson,same 6323
        ret = pmic_config_interface(0x256,0x1,0x1,8); // [8:8]: VSYS_VSLEEP_EN; Johnson,JT
        ret = pmic_config_interface(0x23C,0x1,0x1,1); // [1:1]: VSYS_VOSEL_CTRL; after 0x0256,Johnson
        ret = pmic_config_interface(0x302,0x1,0x3,8); // [9:8]: RG_VPA_CSL; Johnson,same 6323
        ret = pmic_config_interface(0x302,0x1,0x3,14); // [15:14]: RG_VPA_ZX_OS; Johnson,same 6323
        ret = pmic_config_interface(0x310,0x1,0x1,7); // [7:7]: VPA_SFCHG_FEN; Johnson,JT
        ret = pmic_config_interface(0x310,0x1,0x1,15); // [15:15]: VPA_SFCHG_REN; Johnson,JT
        ret = pmic_config_interface(0x326,0x1,0x1,0); // [0:0]: VPA_DLC_MAP_EN; Johnson,same 6323
        ret = pmic_config_interface(0x402,0x1,0x1,0); // [0:0]: VTCXO_LP_SEL; TH,JT
        ret = pmic_config_interface(0x402,0x1,0x1,10); // [10:10]: RG_VTCXO_EN; TH,JT
        ret = pmic_config_interface(0x402,0x0,0x1,11); // [11:11]: VTCXO_ON_CTRL; TH,JT
        ret = pmic_config_interface(0x404,0x1,0x1,0); // [0:0]: VA_LP_SEL; TH,JT
        ret = pmic_config_interface(0x404,0x2,0x3,8); // [9:8]: RG_VA_SENSE_SEL; Check with ricky or KH
        ret = pmic_config_interface(0x500,0x1,0x1,0); // [0:0]: VIO28_LP_SEL; JT,TH
        ret = pmic_config_interface(0x502,0x1,0x1,0); // [0:0]: VUSB_LP_SEL; JT,TH
        ret = pmic_config_interface(0x504,0x1,0x1,0); // [0:0]: VMC_LP_SEL; JT,TH
        ret = pmic_config_interface(0x506,0x1,0x1,0); // [0:0]: VMCH_LP_SEL; JT,TH
        ret = pmic_config_interface(0x508,0x1,0x1,0); // [0:0]: VEMC_3V3_LP_SEL; JT,TH
        ret = pmic_config_interface(0x514,0x1,0x1,0); // [0:0]: RG_STB_SIM1_SIO; CC
        ret = pmic_config_interface(0x516,0x1,0x1,0); // [0:0]: VSIM1_LP_SEL; Chek with Lower power team
        ret = pmic_config_interface(0x518,0x1,0x1,0); // [0:0]: VSIM2_LP_SEL; Chek with Lower power team
        ret = pmic_config_interface(0x524,0x1,0x1,0); // [0:0]: RG_STB_SIM2_SIO; CC
        ret = pmic_config_interface(0x550,0x1,0x1,1); // [1:1]: VRF18_ON_CTRL; JT,TH
        ret = pmic_config_interface(0x552,0x1,0x1,0); // [0:0]: VM_LP_SEL; JT,TH
        ret = pmic_config_interface(0x556,0x1,0x1,0); // [0:0]: VIO18_LP_SEL; JT,TH
        ret = pmic_config_interface(0x756,0x1,0x3,2); // [3:2]: RG_ADC_TRIM_CH6_SEL; Ricky
        ret = pmic_config_interface(0x756,0x1,0x3,4); // [5:4]: RG_ADC_TRIM_CH5_SEL; Ricky
        ret = pmic_config_interface(0x756,0x1,0x3,8); // [9:8]: RG_ADC_TRIM_CH3_SEL; Ricky
        ret = pmic_config_interface(0x756,0x1,0x3,10); // [11:10]: RG_ADC_TRIM_CH2_SEL; Ricky
        ret = pmic_config_interface(0x778,0x1,0x1,15); // [15:15]: RG_VREF18_ENB_MD; JT confirmed with ZF


        /* GPIO5 WORKAROUND */
        ret = pmic_config_interface(0x416,0x1,0x1, 7); // [7:7]: RG_VCN33_EN_BT           
        ret = pmic_config_interface(0x418,0x1,0x1, 12); // [12:12]: RG_VCN33_EN_WIFI
        ret = pmic_config_interface(0x420,0x1,0x1, 1); // [1:1]: VCN33_LP_SET
        ret = pmic_config_interface(0x512,0x1,0x1, 14); // [14:14]: RG_VCN18_EN
        ret = pmic_config_interface(0x512,0x1,0x1, 1); // [1:1]: VCN18_LP_SET

	/* External Buck WORKAROUND */
	ret = pmic_config_interface(0x04C,0x1,0x1, 1); // [1:1]: STRUP_EXT_PMIC_SEL
        //--------------------------------------------------------
}

void PMIC_CUSTOM_SETTING_V1(void)
{
    #if defined(CONFIG_MTK_FPGA)
    #else    
  //    pmu_drv_tool_customization_init(); //DCT
    #endif
}


//==============================================================================
// Dump all LDO status 
//==============================================================================
void dump_ldo_status_read_debug(void)
{
    int i,j;
    int en=0;
    int voltage_reg=0;
    int voltage=0;
    const int *pVoltage;

    printk("********** BUCK/LDO status dump [1:ON,0:OFF]**********\n");

    for (i = 0; i < ARRAY_SIZE(mtk_bucks); i++) 
    {
        if(mtk_bucks[i].qi_en_reg!=0)
        {
            en=pmic_get_register_value(mtk_bucks[i].qi_en_reg);
        }
        else
        {
            en=-1;
        }

        if(mtk_bucks[i].qi_vol_reg!=0)
        {
            voltage_reg=pmic_get_register_value(mtk_bucks[i].qi_vol_reg);
            voltage=mtk_bucks[i].desc.min_uV+mtk_bucks[i].desc.uV_step*voltage_reg;
        }
        else
        {
            voltage_reg=-1;
            voltage=-1;
        }
    }   

    for (i = 0; i < ARRAY_SIZE(mtk_ldos); i++) 
    {
        if(mtk_ldos[i].en_reg!=0)
        {
            en=pmic_get_register_value(mtk_ldos[i].en_reg);
        }
        else
        {
            en=-1;
        }

        if(mtk_ldos[i].desc.n_voltages!=1)
        {
            if(mtk_ldos[i].vol_reg!=0)
            {
                voltage_reg=pmic_get_register_value(mtk_ldos[i].vol_reg);
                if(mtk_ldos[i].pvoltages!=NULL)
                {
                    pVoltage=(const int*)mtk_ldos[i].pvoltages;
                    voltage=pVoltage[voltage_reg];
                }
                else
                {
                    voltage=mtk_ldos[i].desc.min_uV+mtk_ldos[i].desc.uV_step*voltage_reg;
                }
            }
            else
            {
                voltage_reg=-1;
                voltage=-1;
            }
        }
        else
        {
            pVoltage=(const int*)mtk_ldos[i].pvoltages;
            voltage=pVoltage[0];
        }
    }   


    for (i = 0; i < ARRAY_SIZE(interrupts); i++) 
    {
        for(j=0;j<PMIC_INT_WIDTH;j++)
        {

            PMICLOG("[PMIC_INT][%s] interrupt issue times: %d\n", interrupts[i].interrupts[j].name,
                interrupts[i].interrupts[j].times);
        }
    }   

    PMICLOG("Power Good Status=0x%x.\n", upmu_get_reg_value(MT6350_PGSTATUS));
    PMICLOG("OC Status0=0x%x.\n", upmu_get_reg_value(MT6350_OCSTATUS0));
    PMICLOG("OC Status1=0x%x.\n", upmu_get_reg_value(MT6350_OCSTATUS1));
    PMICLOG("Regulator Status0=0x%x.\n", upmu_get_reg_value(MT6350_EN_STATUS0));
    PMICLOG("Regulator Status1=0x%x.\n", upmu_get_reg_value(MT6350_EN_STATUS1));    
    PMICLOG("Thermal Status=0x%x.\n", upmu_get_reg_value(MT6350_STRUP_CON5));
}

static int proc_utilization_show(struct seq_file *m, void *v)
{
    int i,j;
    int en=0;
    int voltage_reg=0;
    int voltage=0;
    const int *pVoltage;

    seq_printf(m,"********** BUCK/LDO status dump [1:ON,0:OFF]**********\n");

    for (i = 0; i < ARRAY_SIZE(mtk_bucks); i++) 
    {
        if(mtk_bucks[i].qi_en_reg!=0)
        {
            en=pmic_get_register_value(mtk_bucks[i].qi_en_reg);
        }
        else
        {
            en=-1;
        }

        if(mtk_bucks[i].qi_vol_reg!=0)
        {
            voltage_reg=pmic_get_register_value(mtk_bucks[i].qi_vol_reg);
            voltage=mtk_bucks[i].desc.min_uV+mtk_bucks[i].desc.uV_step*voltage_reg;
        }
        else
        {
            voltage_reg=-1;
            voltage=-1;
        }
        seq_printf(m,"%s   status:%d     voltage:%duv    voltage_reg:%d\n",mtk_bucks[i].desc.name,en,voltage,voltage_reg);
    }   

    for (i = 0; i < ARRAY_SIZE(mtk_ldos); i++) 
    {
        if(mtk_ldos[i].en_reg!=0)
        {
            en=pmic_get_register_value(mtk_ldos[i].en_reg);
        }
        else
        {
            en=-1;
        }

        if(mtk_ldos[i].desc.n_voltages!=1)
        {
            if(mtk_ldos[i].vol_reg!=0)
            {
                voltage_reg=pmic_get_register_value(mtk_ldos[i].vol_reg);
                if(mtk_ldos[i].pvoltages!=NULL)
                {
                    pVoltage=(const int*)mtk_ldos[i].pvoltages;
                    voltage=pVoltage[voltage_reg];
                }
                else
                {
                    voltage=mtk_ldos[i].desc.min_uV+mtk_ldos[i].desc.uV_step*voltage_reg;
                }
            }
            else
            {
                voltage_reg=-1;
                voltage=-1;
            }
        }
        else
        {
            pVoltage=(const int*)mtk_ldos[i].pvoltages;
            voltage=pVoltage[0];
        }
        seq_printf(m,"%s   status:%d     voltage:%duv    voltage_reg:%d\n",mtk_ldos[i].desc.name,en,voltage,voltage_reg);
    }   

    for (i = 0; i < ARRAY_SIZE(interrupts); i++) 
    {
        for(j=0;j<PMIC_INT_WIDTH;j++)
        {

            seq_printf(m,"[PMIC_INT][%s] interrupt issue times: %d\n", interrupts[i].interrupts[j].name,
                interrupts[i].interrupts[j].times);
        }
    }   
    seq_printf(m,"Power Good Status=0x%x.\n", upmu_get_reg_value(MT6350_PGSTATUS));
    seq_printf(m,"OC Status0=0x%x.\n", upmu_get_reg_value(MT6350_OCSTATUS0));
    seq_printf(m,"OC Status1=0x%x.\n", upmu_get_reg_value(MT6350_OCSTATUS1));
    seq_printf(m,"Regulator Status0=0x%x.\n", upmu_get_reg_value(MT6350_EN_STATUS0));
    seq_printf(m,"Regulator Status1=0x%x.\n", upmu_get_reg_value(MT6350_EN_STATUS1));    
    seq_printf(m,"Thermal Status=0x%x.\n", upmu_get_reg_value(MT6350_STRUP_CON5));
    
    return 0;
}

static int proc_utilization_open(struct inode *inode, struct file *file)
{
    return single_open(file, proc_utilization_show, NULL);
}

static const struct file_operations pmic_debug_proc_fops = { 
    .open  = proc_utilization_open, 
    .read  = seq_read,
};

static int proc_dump_register_show(struct seq_file *m, void *v)
{
    int i;

    seq_printf(m,"********** dump PMIC registers**********\n");

   for(i=0;i<=0x0fae;i=i+10)
   {
    seq_printf(m,"Reg[%x]=0x%x Reg[%x]=0x%x Reg[%x]=0x%x Reg[%x]=0x%x Reg[%x]=0x%x\n",
        i,upmu_get_reg_value(i),i+1,upmu_get_reg_value(i+1),i+2,upmu_get_reg_value(i+2),i+3,upmu_get_reg_value(i+3),i+4,upmu_get_reg_value(i+4));




       seq_printf(m,"Reg[%x]=0x%x Reg[%x]=0x%x Reg[%x]=0x%x Reg[%x]=0x%x Reg[%x]=0x%x\n",
            i+5,upmu_get_reg_value(i+5),i+6,upmu_get_reg_value(i+6),i+7,upmu_get_reg_value(i+7),i+8,upmu_get_reg_value(i+8),i+9,upmu_get_reg_value(i+9));
   }
    
    return 0;
}

static int proc_dump_register_open(struct inode *inode, struct file *file)
{
    return single_open(file, proc_dump_register_show, NULL);
}

static const struct file_operations pmic_dump_register_proc_fops = { 
    .open  = proc_dump_register_open, 
    .read  = seq_read,
};

void pmic_debug_init(void)
{
    struct proc_dir_entry *mt_pmic_dir;

    mt_pmic_dir = proc_mkdir("mt_pmic", NULL);
    if (!mt_pmic_dir) {
        PMICLOG("fail to mkdir /proc/mt_pmic\n" );
        return;
    }

    proc_create("dump_ldo_status", S_IRUGO | S_IWUSR, mt_pmic_dir, &pmic_debug_proc_fops);

    proc_create("dump_pmic_reg", S_IRUGO | S_IWUSR, mt_pmic_dir, &pmic_dump_register_proc_fops);
}

static kal_bool pwrkey_detect_flag = KAL_FALSE;
static struct hrtimer pwrkey_detect_timer;
static struct task_struct *pwrkey_detect_thread = NULL;
static DECLARE_WAIT_QUEUE_HEAD(pwrkey_detect_waiter);

#define BAT_MS_TO_NS(x) (x * 1000 * 1000)

enum hrtimer_restart pwrkey_detect_sw_workaround(struct hrtimer *timer)
{
    kal_uint32 val=0;
    pwrkey_detect_flag = KAL_TRUE;
    wake_up_interruptible(&pwrkey_detect_waiter);
    

    return HRTIMER_NORESTART;
}

int pwrkey_detect_sw_thread_handler(void *unused)
{
    ktime_t ktime;
    do {
        ktime = ktime_set(3, BAT_MS_TO_NS(1000));



        wait_event_interruptible(pwrkey_detect_waiter,
                     (pwrkey_detect_flag == KAL_TRUE));

    //PMICLOG("=>charger_hv_detect_sw_workaround \n");
    if (pmic_get_register_value(PMIC_RG_STRUP_75K_CK_PDN)==1)
    {
        pmic_set_register_value(PMIC_RG_STRUP_75K_CK_PDN,0);
    }


        hrtimer_start(&pwrkey_detect_timer, ktime, HRTIMER_MODE_REL);

    } while (!kthread_should_stop());

    return 0;

}



void pwrkey_sw_workaround_init(void)
{
    ktime_t ktime;

    ktime = ktime_set(0, BAT_MS_TO_NS(2000));
    hrtimer_init(&pwrkey_detect_timer, CLOCK_MONOTONIC, HRTIMER_MODE_REL);
    pwrkey_detect_timer.function = pwrkey_detect_sw_workaround;
    hrtimer_start(&pwrkey_detect_timer, ktime, HRTIMER_MODE_REL);

    pwrkey_detect_thread =
        kthread_run(pwrkey_detect_sw_thread_handler, 0,
            "mtk pwrkey_sw_workaround_init");
    
    if (IS_ERR(pwrkey_detect_thread)) {
        PMICLOG("[%s]: failed to create pwrkey_detect_thread thread\n",
                    __func__);
    }

}

#ifdef LOW_BATTERY_PROTECT
//==============================================================================
// low battery protect UT
//==============================================================================
static ssize_t show_low_battery_protect_ut(struct device *dev,struct device_attribute *attr, char *buf)
{
    return sprintf(buf, "%u\n", g_low_battery_level);
}
static ssize_t store_low_battery_protect_ut(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
    char *pvalue = NULL;
    U32 val = 0;
    
    if(buf != NULL && size != 0)
    {
        val = simple_strtoul(buf,&pvalue,16);
        if(val<=2)
        {
            exec_low_battery_callback(val);
        }
    }
    return size;
}
static DEVICE_ATTR(low_battery_protect_ut, 0664, show_low_battery_protect_ut, store_low_battery_protect_ut); //664

//==============================================================================
// low battery protect stop
//==============================================================================
static ssize_t show_low_battery_protect_stop(struct device *dev,struct device_attribute *attr, char *buf)
{
    return sprintf(buf, "%u\n", g_low_battery_stop);
}
static ssize_t store_low_battery_protect_stop(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
    char *pvalue = NULL;
    U32 val = 0;
    
    if(buf != NULL && size != 0)
    {
        val = simple_strtoul(buf,&pvalue,16);
        if( (val!=0) && (val!=1) )
            val=0;
        g_low_battery_stop = val;
    }
    return size;
}
static DEVICE_ATTR(low_battery_protect_stop, 0664, show_low_battery_protect_stop, store_low_battery_protect_stop); //664

//==============================================================================
// low battery protect level
//==============================================================================
static ssize_t show_low_battery_protect_level(struct device *dev,struct device_attribute *attr, char *buf)
{
    return sprintf(buf, "%u\n", g_low_battery_level);
}
static ssize_t store_low_battery_protect_level(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
    return size;
}
static DEVICE_ATTR(low_battery_protect_level, 0664, show_low_battery_protect_level, store_low_battery_protect_level); //664
#endif

#ifdef BATTERY_OC_PROTECT
//==============================================================================
// battery OC protect UT
//==============================================================================
static ssize_t show_battery_oc_protect_ut(struct device *dev,struct device_attribute *attr, char *buf)
{
    return sprintf(buf, "%u\n", g_battery_oc_level);
}
static ssize_t store_battery_oc_protect_ut(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
    char *pvalue = NULL;
    U32 val = 0;
    
    if(buf != NULL && size != 0)
    {
        val = simple_strtoul(buf,&pvalue,16);
        if(val<=1)
        {
            exec_battery_oc_callback(val);
        }
    }
    return size;
}
static DEVICE_ATTR(battery_oc_protect_ut, 0664, show_battery_oc_protect_ut, store_battery_oc_protect_ut); //664

//==============================================================================
// battery OC protect stop
//==============================================================================
static ssize_t show_battery_oc_protect_stop(struct device *dev,struct device_attribute *attr, char *buf)
{
    return sprintf(buf, "%u\n", g_battery_oc_stop);
}
static ssize_t store_battery_oc_protect_stop(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
    char *pvalue = NULL;
    U32 val = 0;
    
    if(buf != NULL && size != 0)
    {
        val = simple_strtoul(buf,&pvalue,16);
        if( (val!=0) && (val!=1) )
            val=0;
        g_battery_oc_stop = val;
    }
    return size;
}
static DEVICE_ATTR(battery_oc_protect_stop, 0664, show_battery_oc_protect_stop, store_battery_oc_protect_stop); //664

//==============================================================================
// battery OC protect level
//==============================================================================
static ssize_t show_battery_oc_protect_level(struct device *dev,struct device_attribute *attr, char *buf)
{
    return sprintf(buf, "%u\n", g_battery_oc_level);
}
static ssize_t store_battery_oc_protect_level(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
   
    return size;
}
static DEVICE_ATTR(battery_oc_protect_level, 0664, show_battery_oc_protect_level, store_battery_oc_protect_level); //664
#endif

#ifdef BATTERY_PERCENT_PROTECT
//==============================================================================
// battery percent protect UT
//==============================================================================
static ssize_t show_battery_percent_protect_ut(struct device *dev,struct device_attribute *attr, char *buf)
{
    return sprintf(buf, "%u\n", g_battery_percent_level);
}
static ssize_t store_battery_percent_protect_ut(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
    char *pvalue = NULL;
    U32 val = 0;
    
    
    if(buf != NULL && size != 0)
    {
        val = simple_strtoul(buf,&pvalue,16);
        if(val<=1)
        {
            exec_battery_percent_callback(val);
        }
        else
        {
        }
    }
    return size;
}
static DEVICE_ATTR(battery_percent_protect_ut, 0664, show_battery_percent_protect_ut, store_battery_percent_protect_ut); //664

//==============================================================================
// battery percent protect stop
//==============================================================================
static ssize_t show_battery_percent_protect_stop(struct device *dev,struct device_attribute *attr, char *buf)
{
    return sprintf(buf, "%u\n", g_battery_percent_stop);
}
static ssize_t store_battery_percent_protect_stop(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
    char *pvalue = NULL;
    U32 val = 0;
    
    if(buf != NULL && size != 0)
    {
        val = simple_strtoul(buf,&pvalue,16);
        if( (val!=0) && (val!=1) )
            val=0;
        g_battery_percent_stop = val;
    }
    return size;
}
static DEVICE_ATTR(battery_percent_protect_stop, 0664, show_battery_percent_protect_stop, store_battery_percent_protect_stop); //664

//==============================================================================
// battery percent protect level
//==============================================================================
static ssize_t show_battery_percent_protect_level(struct device *dev,struct device_attribute *attr, char *buf)
{
    return sprintf(buf, "%u\n", g_battery_percent_level);
}
static ssize_t store_battery_percent_protect_level(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
    return size;
}
static DEVICE_ATTR(battery_percent_protect_level, 0664, show_battery_percent_protect_level, store_battery_percent_protect_level); //664
#endif

#ifdef DLPT_FEATURE_SUPPORT
//==============================================================================
// DLPT UT
//==============================================================================
static ssize_t show_dlpt_ut(struct device *dev,struct device_attribute *attr, char *buf)
{
    return sprintf(buf, "%u\n", g_dlpt_val);
}
static ssize_t store_dlpt_ut(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
    char *pvalue = NULL;
    U32 val = 0;
    
    if(buf != NULL && size != 0)
    {
        val = simple_strtoul(buf,&pvalue,10);
        
        exec_dlpt_callback(val);        
    }
    return size;
}
static DEVICE_ATTR(dlpt_ut, 0664, show_dlpt_ut, store_dlpt_ut); //664

//==============================================================================
// DLPT stop
//==============================================================================
static ssize_t show_dlpt_stop(struct device *dev,struct device_attribute *attr, char *buf)
{
    return sprintf(buf, "%u\n", g_dlpt_stop);
}
static ssize_t store_dlpt_stop(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
    char *pvalue = NULL;
    U32 val = 0;
    
    if(buf != NULL && size != 0)
    {
        val = simple_strtoul(buf,&pvalue,16);
        if( (val!=0) && (val!=1) )
            val=0;
        g_dlpt_stop = val;
    }
    return size;
}
static DEVICE_ATTR(dlpt_stop, 0664, show_dlpt_stop, store_dlpt_stop); //664

//==============================================================================
// DLPT level
//==============================================================================
static ssize_t show_dlpt_level(struct device *dev,struct device_attribute *attr, char *buf)
{
    return sprintf(buf, "%u\n", g_dlpt_val);
}
static ssize_t store_dlpt_level(struct device *dev,struct device_attribute *attr, const char *buf, size_t size)
{
   
    return size;
}
static DEVICE_ATTR(dlpt_level, 0664, show_dlpt_level, store_dlpt_level); //664
#endif

//==============================================================================
// system function 
//==============================================================================
int g_plug_out_status = 0;

static int detect_battery_plug_out_status(void)
{
	kal_int32 plug_out_1, plug_out_2, ret;

	plug_out_1=pmic_get_register_value(PMIC_STRUP_PWROFF_PREOFF_EN);
	plug_out_2=pmic_get_register_value(PMIC_STRUP_PWROFF_SEQ_EN);

	if (plug_out_1 == 0 && plug_out_2 == 0)
		g_plug_out_status = 1;
	else
		g_plug_out_status = 0;	
}

int get_battery_plug_out_status(void)
{
	return g_plug_out_status;
}


static int pmic_mt_probe(struct platform_device *dev)
{
    int ret_device_file = 0,i;    

    detect_battery_plug_out_status();

    //pmic initial setting
    PMIC_INIT_SETTING_V1();
    PMIC_CUSTOM_SETTING_V1();

//#if defined(CONFIG_MTK_FPGA)
#if 0
    PMICLOG("[PMIC_EINT_SETTING] disable when CONFIG_MTK_FPGA\n");
#else
    //PMIC Interrupt Service
    pmic_thread_handle = kthread_create(pmic_thread_kthread, (void *) NULL, "pmic_thread");
    if (IS_ERR(pmic_thread_handle)) 
    {
        pmic_thread_handle = NULL;
        PMICLOG("[pmic_thread_kthread_mt6350] creation fails\n");        
    }
    else
    {
        wake_up_process(pmic_thread_handle);
    } 
    PMIC_EINT_SETTING();
#endif


    mtk_regulator_init(dev);

#ifdef LOW_BATTERY_PROTECT
    low_battery_protect_init();
#endif

#ifdef BATTERY_OC_PROTECT
    battery_oc_protect_init();
#endif

#ifdef BATTERY_PERCENT_PROTECT
    bat_percent_notify_init();
#endif

#ifdef DLPT_FEATURE_SUPPORT
    dlpt_notify_init();
#endif

#if 0
    pmic_set_register_value(PMIC_AUXADC_CK_AON,1); 
    pmic_set_register_value(PMIC_RG_CLKSQ_EN_AUX_AP_MODE,1);
    PMICLOG("[PMIC] auxadc 26M test : Reg[0x%x]=0x%x, Reg[0x%x]=0x%x\n", 
            MT6328_AUXADC_CON0, upmu_get_reg_value(MT6328_AUXADC_CON0),
            MT6328_TOP_CLKSQ, upmu_get_reg_value(MT6328_TOP_CLKSQ)          
            );
#endif

    pmic_debug_init();

    pmic_ftm_init();

    //EM BUCK voltage & Status
    for (i = 0; i < ARRAY_SIZE(mtk_bucks); i++) 
    {
        //PMICLOG("[PMIC] register buck id=%d \n",i);
        ret_device_file = device_create_file(&(dev->dev),&mtk_bucks[i].en_att);
        ret_device_file = device_create_file(&(dev->dev),&mtk_bucks[i].voltage_att);
    }   

    //EM ldo voltage & Status
    for (i = 0; i < ARRAY_SIZE(mtk_ldos); i++) 
    {
        //PMICLOG("[PMIC] register ldo id=%d \n",i);
        ret_device_file = device_create_file(&(dev->dev),&mtk_ldos[i].en_att);
        ret_device_file = device_create_file(&(dev->dev),&mtk_ldos[i].voltage_att);
    }   

    ret_device_file = device_create_file(&(dev->dev), &dev_attr_pmic_access);
    ret_device_file = device_create_file(&(dev->dev), &dev_attr_pmic_dvt);

#ifdef LOW_BATTERY_PROTECT
    ret_device_file = device_create_file(&(dev->dev), &dev_attr_low_battery_protect_ut);
    ret_device_file = device_create_file(&(dev->dev), &dev_attr_low_battery_protect_stop);
    ret_device_file = device_create_file(&(dev->dev), &dev_attr_low_battery_protect_level);
#endif  

#ifdef BATTERY_OC_PROTECT
    ret_device_file = device_create_file(&(dev->dev), &dev_attr_battery_oc_protect_ut);
    ret_device_file = device_create_file(&(dev->dev), &dev_attr_battery_oc_protect_stop);
    ret_device_file = device_create_file(&(dev->dev), &dev_attr_battery_oc_protect_level);
#endif  

#ifdef BATTERY_PERCENT_PROTECT
    ret_device_file = device_create_file(&(dev->dev), &dev_attr_battery_percent_protect_ut);
    ret_device_file = device_create_file(&(dev->dev), &dev_attr_battery_percent_protect_stop);
    ret_device_file = device_create_file(&(dev->dev), &dev_attr_battery_percent_protect_level);
#endif  

#ifdef DLPT_FEATURE_SUPPORT
    ret_device_file = device_create_file(&(dev->dev), &dev_attr_dlpt_ut);
    ret_device_file = device_create_file(&(dev->dev), &dev_attr_dlpt_stop);
    ret_device_file = device_create_file(&(dev->dev), &dev_attr_dlpt_level);
#endif  

    //pwrkey_sw_workaround_init();
    return 0;
}

static int pmic_mt_remove(struct platform_device *dev)
{
    return 0;
}

static void pmic_mt_shutdown(struct platform_device *dev)
{
}

static int pmic_mt_suspend(struct platform_device *dev, pm_message_t state)
{
#ifdef LOW_BATTERY_PROTECT
    lbat_min_en_setting(0);
    lbat_max_en_setting(0);
#endif
    
#ifdef BATTERY_OC_PROTECT
    bat_oc_h_en_setting(0);
    bat_oc_l_en_setting(0);
#endif

    return 0;
}

static int pmic_mt_resume(struct platform_device *dev)
{
#ifdef LOW_BATTERY_PROTECT
    lbat_min_en_setting(0);
    lbat_max_en_setting(0);
    mdelay(1);

    if(g_low_battery_level==1)
    {
        lbat_min_en_setting(1);
        lbat_max_en_setting(1);
    }
    else if(g_low_battery_level==2)
    {
        //lbat_min_en_setting(0);
        lbat_max_en_setting(1);
    }
    else //0
    {
        lbat_min_en_setting(1);
        //lbat_max_en_setting(0);
    }
#endif
    
#ifdef BATTERY_OC_PROTECT
    bat_oc_h_en_setting(0);
    bat_oc_l_en_setting(0);
    mdelay(1);

    if(g_battery_oc_level==1)
        bat_oc_h_en_setting(1);
    else
        bat_oc_l_en_setting(1);
#endif

    return 0;
}

struct platform_device pmic_mt_device = {
    .name   = "mt-pmic",
    .id        = -1,
};

static struct platform_driver pmic_mt_driver = {
    .probe        = pmic_mt_probe,
    .remove       = pmic_mt_remove,
    .shutdown     = pmic_mt_shutdown,
    //#ifdef CONFIG_PM
    .suspend      = pmic_mt_suspend,
    .resume       = pmic_mt_resume,
    //#endif
    .driver       = {
        .name = "mt-pmic",
    },
};

//==============================================================================
// PMIC mudule init/exit
//==============================================================================
static int __init pmic_mt_init(void)
{
    int ret;

    wake_lock_init(&pmicThread_lock, WAKE_LOCK_SUSPEND, "pmicThread_lock_mt6328 wakelock");
    wake_lock_init(&bat_percent_notify_lock, WAKE_LOCK_SUSPEND,"bat_percent_notify_lock wakelock");

#ifdef DLPT_FEATURE_SUPPORT
    wake_lock_init(&dlpt_notify_lock, WAKE_LOCK_SUSPEND,"dlpt_notify_lock wakelock");
#endif //#ifdef DLPT_FEATURE_SUPPORT

    // PMIC device driver register
    ret = platform_device_register(&pmic_mt_device);
    if (ret) {
        PMICLOG("****[pmic_mt_init] Unable to device register(%d)\n", ret);
        return ret;
    }
    ret = platform_driver_register(&pmic_mt_driver);
    if (ret) {
        PMICLOG("****[pmic_mt_init] Unable to register driver (%d)\n", ret);
        return ret;
    }

#if 0//#ifdef CONFIG_HAS_EARLYSUSPEND
    //register_early_suspend(&pmic_early_suspend_desc);
#endif

    pmic_auxadc_init();

    return 0;
}

static void __exit pmic_mt_exit (void)
{}

fs_initcall(pmic_mt_init);

//module_init(pmic_mt_init);
module_exit(pmic_mt_exit);

MODULE_AUTHOR("WY Chuang");
MODULE_DESCRIPTION("MT PMIC Device Driver");
MODULE_LICENSE("GPL");
